from typing import Optional

import numpy as np
import torch


@torch.no_grad()
def dale_(
        tensor: torch.Tensor,
        inh_ratio: float = 0.5,
        rho: float = 0.99,
        inh_first: bool = True,
        seed: Optional[int] = None
) -> torch.Tensor:
    """
    Initialize a tensor with Dale's law. As default it is half excitatory, half inhibitory. The connections in the
    tensor are considered as i (row) -> j (col).

    :param tensor: tensor to initialize. This tensor will be modified inplace.
    :type tensor: torch.Tensor
    :param inh_ratio: ratio of inhibitory connections. Must be between 0 and 1. Default is 0.5.
    :type inh_ratio: float
    :param rho: The connectivity ratio. Must be between 0 and 1. If rho = 1, the tensor will be fully connected.
        Default is 0.99.
    :type rho: float
    :param inh_first: If True, the inhibitory neurons will be in the first half of the tensor. If False, the neurons
        will be shuffled. Default is True.
    :type inh_first: bool
    :param seed: seed for the random number generator. If None, the seed is not set.
    :type seed: Optional[int]

    :return: The initialized tensor.
    :rtype: torch.Tensor
    """
    assert tensor.ndimension() == 2, "tensor must be 2 dimensional."
    assert 0 <= inh_ratio <= 1, "inh_ratio must be between 0 and 1."
    assert 0 <= rho <= 1, "rho must be between 0 and 1."
    rn_gen = torch.Generator()
    if seed is not None:
        rn_gen.manual_seed(seed)
    N, M = tensor.shape
    i, j = torch.triu_indices(N, M)
    N_0 = int((1 - rho) * len(i))  # Number of zero values
    values_upper = torch.cat(
        (torch.zeros(N_0), torch.normal(0, (1 / np.sqrt(N * rho * (1 - rho))), (len(i) - N_0,), generator=rn_gen))
    ).to(tensor.device)
    values_lower = torch.cat(
        (torch.zeros(N_0), torch.normal(0, (1 / np.sqrt(N * rho * (1 - rho))), (len(i) - N_0,), generator=rn_gen))
    ).to(tensor.device)
    values_upper_rn_indexes = torch.randperm(len(values_upper), generator=rn_gen)
    values_lower_rn_indexes = torch.randperm(len(values_lower), generator=rn_gen)
    tensor[i, j] = values_upper[values_upper_rn_indexes]
    tensor[j, i] = values_lower[values_lower_rn_indexes]
    tensor = torch.abs(tensor)
    if inh_first:
        inh_indexes = torch.arange(int(N * inh_ratio))
    else:
        inh_indexes = torch.randperm(N, generator=rn_gen)[:int(N * inh_ratio)]
    tensor[inh_indexes] *= -1
    return tensor


if __name__ == '__main__':
    from matplotlib import pyplot as plt

    plt.imshow(dale_(torch.empty(100, 100), 0.5, 0.99), cmap="RdBu_r")
    plt.colorbar()
    plt.show()





from typing import List, Iterable, Optional, Sequence, Union, Tuple

import torch


def zero_grad_params(params: Iterable[torch.nn.Parameter]):
    """
    Set the gradient of the parameters to zero.

    :param params: The parameters to set the gradient to zero.
    """
    for p in params:
        p.grad = torch.zeros_like(p).detach()


def compute_jacobian(
        *,
        model: Optional[torch.nn.Module] = None,
        params: Optional[Iterable[torch.nn.Parameter]] = None,
        x: Optional[torch.Tensor] = None,
        y: Optional[torch.Tensor] = None,
        strategy: str = "slow",
):
    """
    Compute the jacobian of the model with respect to the parameters.

    # TODO: check https://medium.com/@monadsblog/pytorch-backward-function-e5e2b7e60140
    # TODO: see https://pytorch.org/tutorials/beginner/blitz/autograd_tutorial.html#sphx-glr-beginner-blitz-autograd-tutorial-py

    :param model: The model to compute the jacobian.
    :param params: The parameters to compute the jacobian with respect to. If None, compute the jacobian
        with respect to all the parameters of the model.
    :param x: The input to compute the jacobian. If None, use y instead.
    :param y: The output to compute the jacobian. If None, use x instead.
    :param strategy: The strategy to use to compute the jacobian. Can be "slow" or "fast". At this time the only
        strategy implemented is "slow".

    :return: The jacobian.
    """
    if params is None:
        assert model is not None, "If params is None, model must be provided."
        params = model.parameters()
    zero_grad_params(params)

    if y is not None:
        if strategy.lower() == "fast":
            y.backward(torch.ones_like(y))
            jacobian = [p.grad.view(-1) for p in params]
        elif strategy.lower() == "slow":
            jacobian = [[] for _ in range(len(list(params)))]
            grad_outputs = torch.eye(y.shape[-1], device=y.device)
            for i in range(y.shape[-1]):
                zero_grad_params(params)
                y.backward(grad_outputs[i], retain_graph=True)
                for p_idx, param in enumerate(params):
                    jacobian[p_idx].append(param.grad.view(-1).detach().clone())
            jacobian = [torch.stack(jacobian[i], dim=-1).T for i in range(len(list(params)))]
        else:
            raise ValueError(f"Unsupported strategy: {strategy}")
    elif x is not None:
        jacobian = torch.autograd.functional.jacobian(model, x, params)
    else:
        raise ValueError("Either x or y must be provided.")
    return jacobian


def dy_dw_local(
        y: torch.Tensor,
        params: Sequence[torch.nn.Parameter],
        grad_outputs: Optional[torch.Tensor] = None,
        retain_graph: bool = True,
        allow_unused: bool = True,
) -> List[torch.Tensor]:
    """
    Compute the derivative of z with respect to the parameters using torch.autograd.grad. If a parameter not
    requires grad, the derivative is set to zero.

    :param y: The tensor to compute the derivative.
    :type y: torch.Tensor
    :param params: The parameters to compute the derivative with respect to.
    :type params: Sequence[torch.nn.Parameter]
    :param grad_outputs: The gradient of the output. If None, use a tensor of ones.
    :type grad_outputs: torch.Tensor or None
    :param retain_graph: If True, the graph used to compute the grad will be retained.
    :type retain_graph: bool
    :param allow_unused: If True, allow the computation of the derivative with respect to a parameter that is not
        used in the computation of z.
    :type allow_unused: bool
    :return: The derivative of z with respect to the parameters.
    :rtype: List[torch.Tensor]
    """
    grad_outputs = torch.ones_like(y) if grad_outputs is None else grad_outputs
    grads_local = []
    for param_idx, param in enumerate(params):
        grad = None
        if param.requires_grad:
            grad = torch.autograd.grad(
                y, param,
                grad_outputs=grad_outputs,
                retain_graph=retain_graph,
                allow_unused=allow_unused,
            )[0]
        if grad is None:
            grad = torch.zeros_like(param)
        grads_local.append(grad)
    return grads_local


def vmap(f):
    # TODO: replace by torch.vmap when it is available
    def wrapper(batch_tensor):
        return torch.stack([f(batch_tensor[i]) for i in range(batch_tensor.shape[0])])

    return wrapper


def filter_parameters(
        parameters: Union[Sequence[torch.nn.Parameter], torch.nn.ParameterList],
        requires_grad: bool = True
) -> List[torch.nn.Parameter]:
    """
    Filter the parameters by their requires_grad attribute.

    :param parameters: The parameters to filter.
    :param requires_grad: The value of the requires_grad attribute to filter.

    :return: The filtered parameters.
    """
    return [p for p in parameters if p.requires_grad == requires_grad]


def get_contributing_params(y, top_level=True):
    """
    Get the parameters that contribute to the computation of y.

    Taken from "https://stackoverflow.com/questions/72301628/find-pytorch-model-parameters-that-dont-contribute-to-loss".

    :param y: The tensor to compute the contribution of the parameters.
    :param top_level: Whether y is a top level tensor or not.
    :type top_level: bool
    :return: A generator of the parameters that contribute to the computation of y.
    """
    nf = y.grad_fn.next_functions if top_level else y.next_functions
    for f, _ in nf:
        try:
            yield f.variable
        except AttributeError:
            pass  # node has no tensor
        if f is not None:
            yield from get_contributing_params(f, top_level=False)


def recursive_detach(tensors: Union[torch.Tensor, Tuple[torch.Tensor], List[torch.Tensor]]):
    if isinstance(tensors, tuple):
        out = tuple([recursive_detach(o) for o in tensors])
    elif isinstance(tensors, list):
        out = [recursive_detach(o) for o in tensors]
    elif isinstance(tensors, dict):
        out = {k: recursive_detach(v) for k, v in tensors.items()}
    elif isinstance(tensors, torch.Tensor):
        out = tensors.detach()
    else:
        out = tensors
    return out


def recursive_detach_(tensors: Union[torch.Tensor, Tuple[torch.Tensor], List[torch.Tensor]]):
    if isinstance(tensors, tuple):
        out = tuple([recursive_detach_(o) for o in tensors])
    elif isinstance(tensors, list):
        out = [recursive_detach_(o) for o in tensors]
    elif isinstance(tensors, dict):
        out = {k: recursive_detach_(v) for k, v in tensors.items()}
    elif isinstance(tensors, torch.Tensor):
        out = tensors.detach_()
    else:
        out = tensors
    return out



import warnings
from collections import defaultdict
from typing import Optional, Sequence, Union, Dict, Callable, List, Tuple, Mapping

import numpy as np
import torch
from torch.utils.hooks import RemovableHandle

from .bptt import BPTT
from ..transforms.base import to_tensor
from ..utils import (
    list_insert_replace_at,
    zero_grad_params,
    recursive_detach,
    recursive_detach_,
    unpack_out_hh,
    format_pred_batch,
    dy_dw_local,
)


class TBPTT(BPTT):
    """
    Truncated Backpropagation Through Time (TBPTT) algorithm.
    """
    def __init__(
            self,
            *,
            params: Optional[Sequence[torch.nn.Parameter]] = None,
            layers: Optional[Union[Sequence[torch.nn.Module], torch.nn.Module]] = None,
            output_layers: Optional[Union[Sequence[torch.nn.Module], torch.nn.Module]] = None,
            optimizer: Optional[torch.optim.Optimizer] = None,
            criterion: Optional[Union[Dict[str, Union[torch.nn.Module, Callable]], torch.nn.Module, Callable]] = None,
            backward_time_steps: Optional[int] = None,
            optim_time_steps: Optional[int] = None,
            **kwargs
    ):
        """
        Constructor for TBPTT class.

        :param params: The parameters to optimize. If None, the parameters of the model's trainer will be used.
        :type params: Optional[Sequence[torch.nn.Parameter]]
        :param layers: The layers to apply the TBPTT algorithm to. If None, the layers of the model's trainer will be used.
        :type layers: Optional[Union[Sequence[torch.nn.Module], torch.nn.Module]]
        :param output_layers: The layers to use as output layers. If None, the output layers of the model's trainer will be used.
        :type output_layers: Optional[Union[Sequence[torch.nn.Module], torch.nn.Module]]
        :param optimizer: The optimizer to use.
        :type optimizer: Optional[torch.optim.Optimizer]
        :param criterion: The criterion to use.
        :type criterion: Optional[Union[Dict[str, Union[torch.nn.Module, Callable]], torch.nn.Module, Callable]]
        :param backward_time_steps: The number of time steps to use for the backward pass.
        :type backward_time_steps: Optional[int]
        :param optim_time_steps: The number of time steps to use for the optimizer step.
        :type optim_time_steps: Optional[int]
        :param kwargs: Additional keyword arguments.

        :keyword float auto_backward_time_steps_ratio: The ratio of the number of time steps to use for the backward pass.
                                Defaults to 0.1.
        :keyword float auto_optim_time_steps_ratio: The ratio of the number of time steps to use for the optimizer step.
                                Defaults to 0.1.
        :keyword float alpha: The alpha value to use for the exponential moving average of the gradients.
        :keyword float grad_norm_clip_value: The value to clip the gradients norm to. This parameter is used to
                        normalize the gradients of the parameters in order to help the convergence and avoid
                        overflowing. Defaults to 1.0.
        :keyword float nan: The value to use to replace the NaN values in the gradients. Defaults to 0.0.
        :keyword float posinf: The value to use to replace the inf values in the gradients. Defaults to 1.0.
        :keyword float neginf: The value to use to replace the -inf values in the gradients. Defaults to -1.0.

        :raises AssertionError: If auto_backward_time_steps_ratio is not between 0 and 1.
        :raises AssertionError: If auto_optim_time_steps_ratio is not between 0 and 1.

        .. seealso::
            :py:meth:`neurotorch.learning_algorithms.bptt.BPTT.__init__`
        """
        super(TBPTT, self).__init__(params=params, layers=layers, optimizer=optimizer, criterion=criterion, **kwargs)
        self.output_layers = output_layers
        self._hidden_layer_names = []
        self._original_forwards = {}  # {layer_name: (layer, layer.forward)}
        self._auto_set_backward_time_steps = backward_time_steps is None
        self.backward_time_steps = backward_time_steps
        self._auto_backward_time_steps_ratio = kwargs.get("auto_backward_time_steps_ratio", 0.1)
        assert 0 <= self._auto_backward_time_steps_ratio <= 1, "auto_backward_time_steps_ratio must be between 0 and 1"
        self._auto_set_optim_time_steps = optim_time_steps is None
        self.optim_time_steps = optim_time_steps
        self._auto_optim_time_steps_ratio = kwargs.get("auto_optim_time_steps_ratio", self._auto_backward_time_steps_ratio)
        assert 0 <= self._auto_optim_time_steps_ratio <= 1, "auto_optim_time_steps_ratio must be between 0 and 1"
        self._data_n_time_steps = 0
        self._layers_buffer = defaultdict(list)
        self._forwards_decorated = False
        self._optim_counter = 0
        self._grads = []
        self.alpha = kwargs.get("alpha", 0.0)
        self.grad_norm_clip_value = to_tensor(kwargs.get("grad_norm_clip_value", torch.inf))
        self.nan = kwargs.get("nan", 0.0)
        self.posinf = kwargs.get("posinf", 1.0)
        self.neginf = kwargs.get("neginf", -1.0)
        self.forwards_hooks: List[RemovableHandle] = []
        self._use_hooks = kwargs.get("use_hooks", False)

    def initialize_output_layers(self, trainer):
        """
        Initialize the output layers of the optimizer. Try multiple ways to identify the output layers if those are not
        provided by the user.

        :Note: Must be called before :meth:`initialize_output_params`.

        :param trainer: The trainer object.
        :return: None.
        """
        if not self.output_layers:
            self.output_layers = []
            possible_attrs = ["output_layers", "output_layer"]
            for attr in possible_attrs:
                obj = getattr(trainer.model, attr, [])
                if isinstance(obj, (Sequence, torch.nn.ModuleList)):
                    obj = list(obj)
                elif isinstance(obj, (Mapping, torch.nn.ModuleDict)):
                    obj = list(obj.values())
                elif isinstance(obj, torch.nn.Module):
                    obj = [obj]
                self.output_layers += list(obj)

        if not self.output_layers:
            raise ValueError("Could not find output layers. Please provide them manually.")

    def initialize_layers(self, trainer):
        """
        Initialize the layers of the optimizer. Try multiple ways to identify the output layers if those are not
        provided by the user.

        :param trainer: The trainer object.

        :return: None
        """
        if not self.layers:
            self.layers = []
            possible_attrs = ["input_layers", "input_layer", "hidden_layers", "hidden_layer"]
            for attr in possible_attrs:
                if hasattr(trainer.model, attr):
                    obj = getattr(trainer.model, attr, [])
                    if isinstance(obj, (Sequence, torch.nn.ModuleList)):
                        obj = list(obj)
                    elif isinstance(obj, (Mapping, torch.nn.ModuleDict)):
                        obj = list(obj.values())
                    elif isinstance(obj, torch.nn.Module):
                        obj = [obj]
                    self.layers += list(obj)
        if not self.layers:
            warnings.warn("No hidden layers found. Please provide them manually if you have any.")

    def _grads_zeros_(self):
        self._grads = [torch.zeros_like(p) for p in self.params]

    def start(self, trainer, **kwargs):
        super().start(trainer)
        self.initialize_output_layers(trainer)
        self.initialize_layers(trainer)
        self._initialize_original_forwards()

    def on_batch_begin(self, trainer, **kwargs):
        super().on_batch_begin(trainer)
        self.trainer = trainer
        if trainer.model.training:
            self._data_n_time_steps = self._get_data_time_steps_from_y_batch(
                trainer.current_training_state.y_batch, trainer.current_training_state.x_batch
            )
            self._maybe_update_time_steps()
            self.optimizer.zero_grad()
            self._grads_zeros_()
            self.decorate_forwards()

    def on_batch_end(self, trainer, **kwargs):
        super().on_batch_end(trainer)
        if trainer.model.training:
            for layer_name in self._layers_buffer:
                backward_t = len(self._layers_buffer[layer_name])
                if backward_t > 0:
                    self._backward_at_t(self._data_n_time_steps - 1, backward_t, layer_name)
                    self.optimizer.step()
        self.undecorate_forwards()
        self._layers_buffer.clear()
        self.optimizer.zero_grad()
        self._grads_zeros_()

    def _get_data_time_steps_from_y_batch(
            self,
            y_batch: Union[torch.Tensor, Dict[str, torch.Tensor]],
            x_batch: Union[torch.Tensor, Dict[str, torch.Tensor]],
    ) -> int:
        """
        Get the number of time steps from the y_batch if it has more than 2 dimensions, otherwise from x_batch.

        :param y_batch: The y_batch from the current training state.
        :param x_batch: The x_batch from the current training state.
        :return: The number of time steps.
        :rtype: int
        :raises ValueError: If y_batch and x_batch are not both either torch.Tensor or dict.
        """
        time_steps = None
        if isinstance(y_batch, torch.Tensor):
            if len(y_batch.shape) > 2:
                time_steps = y_batch.shape[1]
        elif isinstance(y_batch, dict):
            if all([len(y.shape) > 2 for y in y_batch.values()]):
                time_steps = max([y.shape[1] for y in y_batch.values()])
        else:
            raise ValueError(f"y_batch must be either a torch.Tensor or a dict, but got {type(y_batch)}")
        if time_steps is None:
            if isinstance(x_batch, torch.Tensor):
                time_steps = x_batch.shape[1]
            elif isinstance(x_batch, dict):
                time_steps = max([x.shape[1] for x in x_batch.values()])
            else:
                raise ValueError(f"x_batch must be either a torch.Tensor or a dict, but got {type(x_batch)}")
        return time_steps

    def _initialize_original_forwards(self):
        """
        Initialize the original forward functions of the layers (not decorated).

        :return: None
        """
        for layer in self.trainer.model.get_all_layers():
            self._original_forwards[layer.name] = (layer, layer.forward)

    def decorate_forwards(self):
        if self.trainer.model.training:
            if not self._forwards_decorated:
                self._initialize_original_forwards()
            self._hidden_layer_names.clear()

            for layer in self.layers:
                self._hidden_layer_names.append(layer.name)
                if self._use_hooks:
                    hook = layer.register_forward_hook(self._hidden_hook, with_kwargs=True)
                    self.forwards_hooks.append(hook)
                else:
                    layer.forward = self._decorate_hidden_forward(layer.forward, layer.name)

            for layer in self.output_layers:
                if self._use_hooks:
                    hook = layer.register_forward_hook(self._output_hook, with_kwargs=True)
                    self.forwards_hooks.append(hook)
                else:
                    layer.forward = self._decorate_forward(layer.forward, layer.name)
            self._forwards_decorated = True

    def undecorate_forwards(self):
        for name, (layer, original_forward) in self._original_forwards.items():
            layer.forward = original_forward
        for hook in self.forwards_hooks:
            hook.remove()
        self.forwards_hooks.clear()
        self._forwards_decorated = False

    def _maybe_update_time_steps(self):
        if self._auto_set_backward_time_steps:
            self.backward_time_steps = max(1, int(self._auto_backward_time_steps_ratio * self._data_n_time_steps))
        if self._auto_set_optim_time_steps:
            self.optim_time_steps = max(1, int(self._auto_optim_time_steps_ratio * self._data_n_time_steps))
        # if self.backward_time_steps != self.optim_time_steps:
        # 	raise NotImplementedError("backward_time_steps != optim_time_steps is not implemented yet")
        if self.backward_time_steps > self.optim_time_steps:
            raise NotImplementedError("backward_time_steps must be lower or equal to optim_time_steps.")

    def _decorate_hidden_forward(self, forward, layer_name: str) -> Callable:
        """
        Decorate the forward method of a layer to detach the hidden state.

        :param forward: The forward method to decorate.
        :param layer_name: The name of the layer.

        :return: The decorated forward method
        """
        def _forward(*args, **kwargs):
            out = forward(*args, **kwargs)
            t, forecasting = kwargs.get("t", None), kwargs.get("forecasting", False)
            if t is None:
                return out
            out_tensor, hh = unpack_out_hh(out)
            return out_tensor, recursive_detach(hh)
        return _forward

    def _decorate_forward(self, forward, layer_name: str):
        def _forward(*args, **kwargs):
            out = forward(*args, **kwargs)
            t, forecasting = kwargs.get("t", None), kwargs.get("forecasting", False)
            if t is None:
                return out
            out_tensor, hh = unpack_out_hh(out)
            list_insert_replace_at(self._layers_buffer[layer_name], t % self.backward_time_steps, out_tensor)
            self._optim_counter += 1
            if len(self._layers_buffer[layer_name]) == self.backward_time_steps:
                self._backward_at_t(t, self.backward_time_steps, layer_name)
                out = recursive_detach(out)
            if self._optim_counter >= self.optim_time_steps:
                self._make_optim_step()
            return out
        return _forward

    def _hidden_hook(self, module, args, kwargs, output) -> None:
        t, forecasting = kwargs.get("t", None), kwargs.get("forecasting", False)
        if t is None:
            return

        out_tensor, hh = unpack_out_hh(output)
        hh = recursive_detach_(hh)
        return

    def _output_hook(self, module, args, kwargs, output) -> None:
        t, forecasting = kwargs.get("t", None), kwargs.get("forecasting", False)
        if t is None:
            return

        layer_name = module.name
        out_tensor, hh = unpack_out_hh(output)
        list_insert_replace_at(self._layers_buffer[layer_name], t % self.backward_time_steps, out_tensor)
        self._optim_counter += 1
        if len(self._layers_buffer[layer_name]) == self.backward_time_steps:
            self._backward_at_t(t, self.backward_time_steps, layer_name)
            output = recursive_detach_(output)
        if self._optim_counter >= self.optim_time_steps:
            self._make_optim_step()
        return

    def _backward_at_t(self, t: int, backward_t: int, layer_name: str):
        y_batch = self._get_y_batch_slice_from_trainer((t + 1) - backward_t, t + 1, layer_name)
        pred_batch = self._get_pred_batch_from_buffer(layer_name)
        batch_loss = self.apply_criterion(pred_batch, y_batch)
        if batch_loss.grad_fn is None:
            # raise ValueError(
            #     f"batch_loss.grad_fn is None. This is probably an internal error. Please report this issue on GitHub."
            # )
            warnings.warn(
                f"batch_loss.grad_fn is None. This is probably an internal error. Please report this issue on GitHub."
            )
            self._layers_buffer[layer_name].clear()
            return

        if np.isclose(self.alpha, 0.0):
            batch_loss.backward()
        else:
            self._compute_decay_grads_(batch_loss)
            self._apply_grads()
        self._clip_grads()
        self._layers_buffer[layer_name].clear()

    def _compute_decay_grads_(self, batch_loss):
        output_grads = dy_dw_local(torch.mean(batch_loss), self.params, retain_graph=True, allow_unused=True)
        with torch.no_grad():
            self._grads = [
                self.alpha * g + torch.nan_to_num(
                    dy_dw.to(g.device),
                    nan=0.0,
                    neginf=-self.grad_norm_clip_value,
                    posinf=self.grad_norm_clip_value,
                )
                for g, dy_dw in zip(self._grads, output_grads)
            ]

    def _apply_grads(self):
        with torch.no_grad():
            for p, g in zip(self.params, self._grads):
                p.grad = g.to(p.device)

    def _clip_grads(self):
        if torch.isfinite(self.grad_norm_clip_value):
            torch.nn.utils.clip_grad_norm_(self.params, self.grad_norm_clip_value)

    def _make_optim_step(self, **kwargs):
        self.optimizer.step()
        self.optimizer.zero_grad()
        zero_grad_params(self.params)
        self._optim_counter = 0

    def _get_y_batch_slice_from_trainer(self, t_first: int, t_last: int, layer_name: str = None):
        """
        Get a slice of the y_batch from the current training state given the first and last time steps. In case
        y_batch is a dict, the slice is applied to all the values of the dict. In case y_batch is a torch.Tensor with
        less than 3 dimensions, the y_batch is stacked along the second dimension t_last - t_first times.

        :param t_first: first time step
        :param t_last: last time step
        :param layer_name: if y_batch is a dict, the slice is applied to the value corresponding to this key
        :return: a slice of the y_batch from the current training state
        """
        y_batch = self.trainer.current_training_state.y_batch.clone()
        if isinstance(y_batch, dict):
            if layer_name is None:
                y_batch = {
                    key: (
                        val[:, t_first:t_last]
                        if len(val.shape) > 2
                        else torch.stack([val for _ in range(t_last - t_first)], dim=1)
                    )
                    for key, val in y_batch.items()
                }
            else:
                y_batch = (
                    y_batch[layer_name][:, t_first:t_last]
                    if len(y_batch[layer_name].shape) > 2
                    else torch.stack([y_batch[layer_name] for _ in range(t_last - t_first)], dim=1)
                )
        else:
            y_batch = (
                y_batch[:, t_first:t_last]
                if len(y_batch.shape) > 2
                else torch.stack([y_batch for _ in range(t_last - t_first)], dim=1)
            )
        return y_batch

    def _get_pred_batch_from_buffer(self, layer_name: str):
        pred_batch = torch.stack(self._layers_buffer[layer_name], dim=1)
        return pred_batch

    def on_optimization_begin(self, trainer, **kwargs):
        y_batch = trainer.current_training_state.y_batch
        pred_batch = format_pred_batch(trainer.current_training_state.pred_batch, y_batch)
        batch_loss = self.apply_criterion(pred_batch, y_batch)
        trainer.update_state_(batch_loss=batch_loss)

    def on_optimization_end(self, trainer, **kwargs):
        super(TBPTT, self).on_optimization_end(trainer)
        self._layers_buffer.clear()

    def close(self, trainer, **kwargs):
        self.undecorate_forwards()
        super(TBPTT, self).close(trainer)

    def extra_repr(self) -> str:
        _repr = f"backward_time_steps: {self.backward_time_steps}, "
        _repr += f"optim_time_steps: {self.optim_time_steps}, "
        _repr += f"alpha: {self.alpha}, "
        _repr += f"grad_norm_clip_value: {self.grad_norm_clip_value}"
        return _repr





from .spike_funcs import (
    SpikeFuncType,
    SpikeFunction,
    SpikeFuncType2Func,
    HeavisideSigmoidApprox,
    HeavisidePhiApprox,
)

from .layers import (
    LayerType,
    BaseLayer,
    Linear,
    LIFLayer,
    ALIFLayer,
    IzhikevichLayer,
    LILayer,
    LayerType2Layer,
    SpyLIFLayer,
    SpyALIFLayer,
    SpyLILayer,
    SpyLIFLayerLPF,
    SpyALIFLayerLPF,
    LIFLayerLPF,
    ALIFLayerLPF,
)


from .base import (
    BaseModel,
)

from .sequential_rnn import (
    SequentialRNN,
)

from .functions import (
    PSigmoid,
)



from typing import List, Optional, Iterable, Union, Sequence

import numpy as np
import torch

from .base_callback import BaseCallback
from ..learning_algorithms.learning_algorithm import LearningAlgorithm


class LinearLRScheduler(BaseCallback):
    """
    This class is a callback that implements a linear learning rate decay. This is useful to decrease the learning rate
    over iterations. The learning rate is decreased linearly from lr_start to lr_end over n_steps iterations.

    :Attributes:
        - **lr_start** (float): The initial learning rate.
        - **lr_end** (float): The final learning rate.
        - **n_steps** (int): The number of steps over which the learning rate is decreased.
        - **lr** (float): The current learning rate.
        - **lr_decay** (float): The learning rate decay per step.
    """
    DEFAULT_LR_START = 1e-3
    DEFAULT_MIN_LR = 1e-12

    def __init__(
            self,
            lr_start: Union[float, Sequence[float]],
            lr_end: Union[float, Sequence[float]],
            n_steps: int,
            optimizer: Optional[torch.optim.Optimizer] = None,
            **kwargs
    ):
        """
        Construcor for the LinearLRScheduler class.

        :param lr_start: The initial learning rate. If a sequence is given, each entry of the sequence is used for each
            parameter group.
        :type lr_start: Union[float, Sequence[float]]
        :param lr_end: The final learning rate. If a sequence is given, each entry of the sequence is used for each
            parameter group.
        :type lr_end: Union[float, Sequence[float]]
        :param n_steps: The number of steps over which the learning rate is decreased.
        :type n_steps: int
        """
        super().__init__()
        self.lr_start = lr_start if isinstance(lr_start, Iterable) else [lr_start]
        self.lr_end = lr_end if isinstance(lr_end, Iterable) else [lr_end]
        self.n_steps = n_steps
        self.lr = self.lr_start
        self.lr_decay = [
            (_lr_start - _lr_end) / self.n_steps
            for _lr_start, _lr_end in zip(self.lr_start, self.lr_end)
        ]
        self.optimizer = optimizer
        self.log_lr_to_history = kwargs.get('log_lr_to_history', True)

    def _init_lr_decay(self):
        assert len(self.lr_start) == len(self.lr_end), \
            "The number of learning rates must match the number of minimum learning rates."
        for i, (_lr_start, _min_lr) in enumerate(zip(self.lr_start, self.lr_end)):
            _decay = (self.lr_start[i] - self.lr_end[i]) / self.n_steps
            if len(self.lr_decay) <= i:
                self.lr_decay.append(_decay)
            if self.lr_decay[i] is None:
                self.lr_decay[i] = _decay

    def start(self, trainer, **kwargs):
        """
        Initialize the learning rate of the optimizer and the :attr:`lr_start` attribute if necessary.

        :param trainer: The trainer object.
        :type trainer: Trainer

        :return: None
        """
        if self.optimizer is None:
            learning_algorithms = trainer.learning_algorithms
            for la in learning_algorithms:
                if isinstance(la, LearningAlgorithm) and hasattr(la, 'optimizer'):
                    self.optimizer = la.optimizer
                    break
            if self.optimizer is None:
                raise ValueError('No optimizer found in the callbacks of the trainer.')
        for i, g in enumerate(self.optimizer.param_groups):
            if len(self.lr_start) <= i:
                self.lr_start.append(g.get("lr", self.DEFAULT_LR_START))
            if self.lr_start[i] is None:
                self.lr_start[i] = g.get("lr", self.DEFAULT_LR_START)

        if len(self.lr_end) < len(self.lr_start):
            for i in range(len(self.lr_start)):
                if len(self.lr_end) <= i:
                    self.lr_end.append(self.DEFAULT_MIN_LR)
                if self.lr_end[i] is None:
                    self.lr_end[i] = self.DEFAULT_MIN_LR
        self._init_lr_decay()

    def on_iteration_end(self, trainer, **kwargs):
        """
        Decrease the learning rate linearly.

        :param trainer: The trainer object.
        :type trainer: Trainer

        :return: None
        """
        if self.log_lr_to_history:
            trainer.update_itr_metrics_state_(**{f'lr_{i}': _lr for i, _lr in enumerate(self.lr)})
        step = trainer.current_training_state.iteration
        self.lr = [
            max(_lr_start - step * self.lr_decay, _lr_end)
            for _lr_start, _lr_end in zip(self.lr_start, self.lr_end)
        ]
        assert len(self.lr) > 0, "No learning rate found."
        if len(self.lr) == 1:
            self.lr = [self.lr[0] for _ in self.optimizer.param_groups]
        assert len(self.lr) == len(self.optimizer.param_groups), \
            "The number of learning rates must match the number of parameter groups."
        for g, lr in zip(self.optimizer.param_groups, self.lr):
            g['lr'] = lr

    def extra_repr(self) -> str:
        return f"lr_start={self.lr_start}, lr_end={self.lr_end}, n_steps={self.n_steps}"


class LRSchedulerOnMetric(BaseCallback):
    """
    Class to schedule the learning rate of the optimizer based on the metric value.
    Each time the metric reach the next value of the schedule, the learning rate is multiplied by the given decay.
    The learning rate is also capped at the given minimum value.

    :Attributes:
        - **metric** (str): The metric to use to schedule the learning rate.
        - **metric_schedule** (Iterable[float]): The schedule of the metric.
        - **minimize_metric** (bool): Whether to minimize the metric or maximize it.
        - **lr_decay** (float): The decay factor to use when the metric reach the next value of the schedule.
        - **min_lr** (float): The minimum learning rate to use.
        - **lr_start** (float): The learning rate to use at the beginning of the training.
        - **lr** (float): The current learning rate.
        - **retain_progress** (bool): If True the current step of the scheduler will only increase when the metric reach
        the next value of the schedule. If False, the current step will increase or decrease depending on the metric.
        - **step** (int): The current step of the scheduler.
    """
    DEFAULT_LR_START = 1e-3
    DEFAULT_MIN_LR = 1e-12

    def __init__(
            self,
            metric: str,
            metric_schedule: Iterable[float],
            *,
            minimize_metric: Optional[bool] = None,
            lr_decay: Optional[Union[float, Sequence[float]]] = None,
            min_lr: Optional[Union[float, Sequence[float]]] = None,
            lr_start: Optional[Union[float, Sequence[float]]] = None,
            retain_progress: bool = True,
            optimizer: Optional[torch.optim.Optimizer] = None,
            **kwargs
    ):
        """
        Initialize the scheduler with the given metric and metric schedule.

        :param metric: The metric to use to schedule the learning rate.
        :type metric: str
        :param metric_schedule: The schedule of the metric.
        :type metric_schedule: Iterable[float]
        :param minimize_metric: Whether to minimize the metric or maximize it. If None, infer from the metric schedule.
        :type minimize_metric: Optional[bool]
        :param lr_decay: The decay factor to use when the metric reach the next value of the schedule. If None, the
        decay is computed automatically as (lr_start - min_lr) / len(metric_schedule).
        :type lr_decay: Optional[Union[float, Sequence[float]]]
        :param min_lr: The minimum learning rate to use.
        :type min_lr: Optional[Union[float, Sequence[float]]]
        :param lr_start: The learning rate to use at the beginning of the training. If None, the learning rate is
        get automatically as the learning rate of the first group of the optimizer.
        :type lr_start: Optional[Union[float, Sequence[float]]]
        :param retain_progress: If True the current step of the scheduler will only increase when the metric reach the
        next value of the schedule. If False, the current step will increase or decrease depending on the metric.
        :type retain_progress: bool
        :param optimizer: The optimizer whose learning rate will be scheduled. If None, the optimizer is get from the
            trainer. Note that in this case the first optimizer of the trainer's callbacks will be used.
        :type optimizer: Optional[torch.optim.Optimizer]

        :param kwargs: The keyword arguments to pass to the BaseCallback.

        :keyword bool log_lr_to_history: Whether to log the learning rate to the training history. Default: True.
        """
        super().__init__(**kwargs)
        self.metric = metric
        self.metric_schedule = np.asarray(metric_schedule)
        self._check_schedule_ascending_or_descending()
        self.minimize_metric = minimize_metric
        self._init_minimize_metric()
        self.lr_decay = lr_decay if isinstance(lr_decay, Sequence) else [lr_decay]
        self.min_lr = min_lr if isinstance(min_lr, Sequence) else [min_lr]
        self.lr_start = lr_start if isinstance(lr_start, Sequence) else [lr_start]
        self.lr = self.lr_start
        self.retain_progress = retain_progress
        self.step = 0
        self.optimizer = optimizer
        self.log_lr_to_history = kwargs.get('log_lr_to_history', True)

    def on_iteration_end(self, trainer, **kwargs):
        """
        Update the learning rate of the optimizer based on the metric value.

        :param trainer: The trainer object.
        :type trainer: Trainer

        :return: None
        """
        last_metric = trainer.training_history[self.metric][-1]
        if self.log_lr_to_history:
            trainer.update_itr_metrics_state_(**{f'lr_{i}': _lr for i, _lr in enumerate(self.lr)})
        self.step = self.update_step(last_metric)
        self.lr = [
            max(_lr_start - self.step * _lr_decay, _min_lr)
            for _lr_start, _lr_decay, _min_lr in zip(self.lr_start, self.lr_decay, self.min_lr)
        ]
        assert len(self.lr) > 0, "No learning rate found."
        if len(self.lr) == 1:
            self.lr = [self.lr[0] for _ in self.optimizer.param_groups]
        assert len(self.lr) == len(self.optimizer.param_groups), \
            "The number of learning rates must match the number of parameter groups."
        for g, lr in zip(self.optimizer.param_groups, self.lr):
            g['lr'] = lr

    def _check_schedule_ascending_or_descending(self):
        sig_diff = np.sign(np.diff(self.metric_schedule))
        all_positive = np.all(sig_diff >= 0)
        all_negative = np.all(sig_diff <= 0)
        if not (all_positive or all_negative):
            raise ValueError('The metric schedule must be ascending or descending.')
        if all_positive:
            self.metric_schedule = np.concatenate(([-np.inf], self.metric_schedule))
        if all_negative:
            self.metric_schedule = np.concatenate(([np.inf], self.metric_schedule))

    def _init_minimize_metric(self):
        if self.minimize_metric is None:
            self.minimize_metric = np.mean(np.diff(self.metric_schedule) <= 0) > 0.5

    def _init_lr_decay(self):
        assert len(self.lr_start) == len(self.min_lr), \
            "The number of learning rates must match the number of minimum learning rates."
        for i, (_lr_start, _min_lr) in enumerate(zip(self.lr_start, self.min_lr)):
            _decay = (self.lr_start[i] - self.min_lr[i]) / len(self.metric_schedule)
            if len(self.lr_decay) <= i:
                self.lr_decay.append(_decay)
            if self.lr_decay[i] is None:
                self.lr_decay[i] = _decay

    def update_step(self, last_metric: float) -> int:
        """
        Update the current step of the scheduler based on the metric value.

        :param last_metric: The last value of the metric.
        :type last_metric: float

        :return: The new step.
        :rtype: int
        """
        last_index = len(self.metric_schedule) - 1
        if self.minimize_metric:
            next_step = last_index - np.argmax((last_metric <= self.metric_schedule)[::-1])
        else:
            next_step = last_index - np.argmax((last_metric >= self.metric_schedule)[::-1])
        if self.retain_progress:
            next_step = max(self.step, next_step)
        self.step = next_step
        return self.step

    def start(self, trainer, **kwargs):
        """
        Initialize the learning rate of the optimizer and the :attr:`lr_start` attribute if necessary.

        :param trainer: The trainer object.
        :type trainer: Trainer

        :return: None
        """
        if self.optimizer is None:
            learning_algorithms = trainer.learning_algorithms
            for la in learning_algorithms:
                if isinstance(la, LearningAlgorithm) and hasattr(la, 'optimizer'):
                    self.optimizer = la.optimizer
                    break
            if self.optimizer is None:
                raise ValueError('No optimizer found in the callbacks of the trainer.')
        for i, g in enumerate(self.optimizer.param_groups):
            if len(self.lr_start) <= i:
                self.lr_start.append(g.get("lr", self.DEFAULT_LR_START))
            if self.lr_start[i] is None:
                self.lr_start[i] = g.get("lr", self.DEFAULT_LR_START)

        if len(self.min_lr) < len(self.lr_start):
            for i in range(len(self.lr_start)):
                if len(self.min_lr) <= i:
                    self.min_lr.append(self.DEFAULT_MIN_LR)
                if self.min_lr[i] is None:
                    self.min_lr[i] = self.DEFAULT_MIN_LR

        self.lr = self.lr_start
        self._init_lr_decay()

    def on_pbar_update(self, trainer, **kwargs) -> dict:
        """
        Return the learning rate to display in the progress bar.

        :param trainer: The trainer object.
        :type trainer: Trainer

        :return: The dictionary to update the progress bar.
        :rtype: dict
        """
        return {}

    def extra_repr(self) -> str:
        return f"metric={self.metric}, minimize={self.minimize_metric}"



import numpy as np
import torch

from ..transforms.base import to_tensor


class WeirdTanh(torch.nn.Module):
    r"""
    Applies the Hyperbolic Tangent (Tanh) function element-wise.

    Tanh is defined as:

    .. math::
        \text{Tanh}(x) = \tanh(x) = \frac{a\exp(\alpha x) - b\exp(-\beta x)} {c\exp(\gamma x) + d\exp(-\delta x)}
    """
    def __init__(
            self,
            a: float = 1.0, b: float = 1.0, c: float = 1.0, d: float = 1.0,
            alpha: float = 1.0, beta: float = 1.0, gamma: float = 1.0, delta: float = 1.0,
    ):
        super().__init__()
        self.a = a
        self.b = b
        self.c = c
        self.d = d
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma
        self.delta = delta

    def forward(self, x):
        numerator = self.a * torch.exp(self.alpha * x) - self.b * torch.exp(-self.beta * x)
        denominator = self.c * torch.exp(self.gamma * x) + self.d * torch.exp(-self.delta * x)
        return numerator / denominator


class PSigmoid(torch.nn.Module):
    r"""
    Applies the Pseudo-Sigmoid function element-wise.

    Pseudo-Sigmoid is defined as:

    .. math::
        \\text{PSigmoid}(x) = \\frac{1}{1 + \\exp(- p \\odot x)}
    """
    def __init__(self, p: float = 1.0, learn_p: bool = True):
        super().__init__()
        self.p_sqrt = torch.nn.Parameter(to_tensor(np.sqrt(p)), requires_grad=learn_p)

    @property
    def p(self):
        return torch.pow(self.p_sqrt, 2)

    def forward(self, x):
        return 1 / (1 + torch.exp(-self.p * x))



import enum
from typing import Any

import torch


class SpikeFuncType(enum.Enum):
    FastSigmoid = 0
    Phi = 1


class SpikeFunction(torch.autograd.Function):
    """
    Implementation of the spike function. The spike function is a differentiable version of the Heaviside function.
    The Heaviside function is defined as the equation :eq:`heaviside`. The backward pass of this function has to be
    an approximation of the derivative of the Heaviside function.

    .. math::
        :label: heaviside

        \\begin{equation}
            H(x, thr) = \\left\\{
            \\begin{matrix}
                1 & \\text{ if } x > thr; \\\\
                0 & \\text{ else}.
            \\end{matrix}
            \\right.
        \\end{equation}

    """

    @staticmethod
    def forward(
            ctx: torch.autograd.function.FunctionCtx,
            inputs: torch.Tensor,
            threshold: torch.Tensor = torch.tensor(1.0),
            gamma: torch.Tensor = torch.tensor(1.0)
    ) -> torch.Tensor:
        """
        The forward pass of the spike function is the Heaviside function. See the heaviside equation.

        :param ctx: The context of the function. It is used to store information for the backward pass. Use the method
            :func:`ctx.save_for_backward` to store information.
        :type ctx: torch.autograd.function.FunctionCtx
        :param inputs: The input tensor.
        :type inputs: torch.Tensor
        :param threshold: The threshold of the spike function.
        :type threshold: torch.Tensor
        :param gamma: The gamma parameter of the spike function. This parameter is used in the backward pass to
            increase the gradient of the spike function. See child classes for more information.
        :type gamma: torch.Tensor

        :return: The output of the spike function.
        :rtype: torch.Tensor
        """
        ctx.save_for_backward(inputs, threshold, gamma)
        out = torch.zeros_like(inputs)
        out[inputs >= threshold] = 1.0
        return out

    @staticmethod
    def backward(ctx: torch.autograd.function.FunctionCtx, grad_outputs):
        """
        In the backward pass we receive a Tensor we need to compute the
        surrogate gradient of the loss with respect to the input.
        Here we use the normalized negative part of a fast sigmoid
        as this was done in Zenke & Ganguli (2018).
        """
        raise NotImplementedError()


# @staticmethod
# def symbolic(g, inputs: torch._C.Value) -> torch._C.Value:
# 	return g.op("SpikeFunction", inputs, g.op("Constant", value_t=torch.tensor(0, dtype=torch.float)))


class HeavisideSigmoidApprox(SpikeFunction):
    """
    Implementation of the spike function. The spike function is a differentiable version of the Heaviside function.
    The Heaviside function is defined in the doc of :class:`SpikeFunction`. The backward pass of this function is the
    first derivative of the fast sigmoid function defined in equation :eq:`fast_sigmoid`. The derivative
    is shown in equation :eq:`fast_sigmoid_derivative` used in Zenke & Ganguli (2018).

    .. math::
        \\begin{equation}
            S(x) = \\frac{1}{1 + e^{-x}}
        \\end{equation}
        :label: fast_sigmoid

    .. math::
        \\begin{equation}
            S'(x) \\approx \\frac{x}{\\left(1 + \\gamma\\vert{x - thr}\\vert\\right)^2}
        \\end{equation}
        :label: fast_sigmoid_derivative

    """

    @staticmethod
    def backward(ctx: torch.autograd.function.FunctionCtx, grad_outputs: torch.Tensor) -> Any:
        """
        The implementation of the equation :eq:`fast_sigmoid_derivative`.

        :param ctx: The context of the function. It is used to retrieve information from the forward pass.
        :type ctx: torch.autograd.function.FunctionCtx
        :param grad_outputs: The gradient of the loss with respect to the output of the forward pass.
        :type grad_outputs: torch.Tensor
        :return: The gradient of the loss with respect to the input of the forward pass.
        """
        inputs, threshold, scale = ctx.saved_tensors
        grad_outputs_clone = grad_outputs.clone()
        inputs_grad = grad_outputs_clone / (scale * torch.abs(inputs - threshold) + 1.0) ** 2
        if ctx.needs_input_grad[1]:
            threshold_grad = -inputs_grad.clone()
        else:
            threshold_grad = None
        return inputs_grad, threshold_grad, None


class HeavisidePhiApprox(SpikeFunction):
    """
    Implementation of the spike function. The spike function is a differentiable version of the Heaviside function.
    The Heaviside function is defined in the doc of :class:`SpikeFunction`. The backward pass of this function is the
    approximation of the heaviside used in :cite:t:`bellec_solution_2020`. This approximation is defined in equation
    :eq:`heaviside_phi_approx`.

    .. math::
        \\begin{equation}
            \\psi_j^t = \\frac{\\gamma_\\text{pd}}{v_{\\text{th}}}
            \\text{max}\\left(0, 1 - \\left\\vert\\frac{v_j^t - A_j^t}{v_\\text{th}}\\right\\vert\\right)
        \\end{equation}
        :label: heaviside_phi_approx

    .. bibliography::

    """
    epsilon = 1e-5

    @staticmethod
    def pseudo_derivative(inputs, threshold, gamma):
        return (gamma / (threshold + HeavisidePhiApprox.epsilon)) * torch.max(
            torch.zeros_like(inputs), 1 - torch.abs((inputs - threshold) / (threshold + HeavisidePhiApprox.epsilon))
        )

    @staticmethod
    def backward(ctx: torch.autograd.function.FunctionCtx, grad_outputs):
        """
        The implementation of the equation :eq:`heaviside_phi_approx`.

        :param ctx: The context of the function. It is used to retrieve information from the forward pass.
        :type ctx: torch.autograd.function.FunctionCtx
        :param grad_outputs: The gradient of the loss with respect to the output of the forward pass.
        :type grad_outputs: torch.Tensor
        :return: The gradient of the loss with respect to the input of the forward pass.
        """
        inputs, threshold, gamma = ctx.saved_tensors
        inputs_grad = grad_outputs.clone() * (gamma / (threshold + HeavisidePhiApprox.epsilon)) * torch.max(
            torch.zeros_like(inputs), 1 - torch.abs((inputs - threshold) / (threshold + HeavisidePhiApprox.epsilon))
        )
        if ctx.needs_input_grad[1]:
            threshold_grad = -inputs_grad.clone()
        else:
            threshold_grad = None
        return inputs_grad, threshold_grad, None


SpikeFuncType2Func = {
    SpikeFuncType.FastSigmoid: HeavisideSigmoidApprox,
    SpikeFuncType.Phi: HeavisidePhiApprox,
}

if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import numpy as np

    th, sc = torch.tensor(1.0), torch.tensor(1.0)
    funcs = {
        "FastSigmoid": HeavisideSigmoidApprox.apply,
        "Bellec": HeavisidePhiApprox.apply
    }
    grads = {
        name: [] for name in funcs
    }
    X = torch.tensor(np.linspace(th - 2, th + 2, num=1_000), requires_grad=False)
    Y = SpikeFunction.apply(X, th, sc)
    for name, func in funcs.items():
        for x_i in X:
            x = torch.tensor(x_i.clone().detach(), requires_grad=True)
            y = func(x, th, sc)
            # y.retain_grad()
            y.backward()
            grads[name].append(x.grad.detach().cpu().numpy())

    plt.plot(X, Y, label="Heaviside")
    for name in funcs:
        plt.plot(X, grads[name], label=name)
    plt.xlabel("V [mV]")
    plt.legend()
    plt.show()



import json
import logging
from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional, Union, Tuple, Iterable

import torch
from torch import nn
from torchvision.transforms import Compose, Lambda
import torch.nn.functional as F

from ..callbacks import CheckpointManager, LoadCheckpointMode
from ..dimension import DimensionLike, SizeTypes, Dimension, DimensionProperty
from ..transforms import to_tensor
from ..transforms.wrappers import CallableToModuleWrapper
from ..transforms.base import IdentityTransform, ToDevice, ToTensor
from ..utils import ravel_compose_transforms, list_of_callable_to_sequential


class NamedModule(torch.nn.Module):
    def __init__(self, name: Optional[str] = None):
        super().__init__()
        self._name_is_set = False
        self.name = name
        self._name_is_default = name is None

    @property
    def name(self) -> str:
        """
        Returns the name of the module. If the name is not set, it will be set to the class name.

        :return: The name of the module.
        """
        if self._name is None:
            return self.__class__.__name__
        return self._name

    @property
    def name_is_set(self) -> bool:
        """
        Returns whether the name of the module has been set.

        :return: Whether the name of the module has been set.
        """
        return self._name_is_set

    @name.setter
    def name(self, name: str):
        """
        Sets the name of the module.

        :param name: The name of the module.
        :return: None
        """
        self._name = name
        if name is not None:
            assert isinstance(name, str), "name must be a string."
            self._name_is_set = True


class SizedModule(NamedModule):

    def __init__(
            self,
            input_size: Optional[SizeTypes] = None,
            output_size: Optional[SizeTypes] = None,
            name: Optional[str] = None,
    ):
        super().__init__(name)
        self.input_size = input_size
        self.output_size = output_size

    @property
    def input_size(self) -> Optional[Dimension]:
        if not hasattr(self, "_input_size"):
            return None
        return self._input_size

    @input_size.setter
    def input_size(self, size: Optional[SizeTypes]):
        self._input_size = self._format_size(size)

    @property
    def output_size(self) -> Optional[Dimension]:
        if not hasattr(self, "_output_size"):
            return None
        return self._output_size

    @output_size.setter
    def output_size(self, size: Optional[SizeTypes]):
        self._output_size = self._format_size(size)

    def _format_size(self, size: Optional[SizeTypes], **kwargs) -> Optional[Dimension]:
        filter_time = kwargs.get("filter_time", getattr(self, "size_filter_time", False))
        # TODO: must accept multiple time dimensions
        if size is not None:
            if isinstance(size, Iterable):
                size = [Dimension.from_int_or_dimension(s) for s in size]
                if filter_time:
                    time_dim_count = len(list(filter(lambda d: d.dtype == DimensionProperty.TIME, size)))
                    assert time_dim_count <= 1, "Size must not contain more than one Time dimension."
                    size = list(filter(lambda d: d.dtype != DimensionProperty.TIME, size))
                if len(size) == 1:
                    size = size[0]
                else:
                    raise ValueError(
                        "Size must be a single dimension or a list of 2 dimensions with a Time one "
                        "if `filter_time` is True."
                    )
            assert isinstance(size, (int, Dimension)), "Size must be an int or Dimension."
            size = Dimension.from_int_or_dimension(size)
        return size


class BaseModel(NamedModule):
    """
    This class is the base class of all models.

    :Attributes:
        - input_sizes: The input sizes of the model.
        - input_transform (torch.nn.ModuleDict): The transforms to apply to the inputs.
        - output_sizes: The output size of the model.
        - output_transform (torch.nn.ModuleDict): The transforms to apply to the outputs.
        - name: The name of the model.
        - checkpoint_folder: The folder where the checkpoints are saved.
        - kwargs: Additional arguments.

    """
    @staticmethod
    def _format_sizes(sizes: Union[Dict[str, DimensionLike], SizeTypes]) -> Dict[str, int]:
        if isinstance(sizes, dict):
            return sizes
        elif isinstance(sizes, list):
            return {
                f"{i}": s
                for i, s in enumerate(sizes)
            }
        else:
            return {
                "0": sizes
            }

    def __init__(
            self,
            input_sizes: Optional[Union[Dict[str, DimensionLike], SizeTypes]] = None,
            output_size: Optional[Union[Dict[str, DimensionLike], SizeTypes]] = None,
            name: str = "BaseModel",
            checkpoint_folder: str = "checkpoints",
            device: torch.device = None,
            input_transform: Union[Dict[str, Callable], List[Callable]] = None,
            output_transform: Union[Dict[str, Callable], List[Callable]] = None,
            **kwargs
    ):
        """
        Constructor of the BaseModel class. This class is the base class of all models.

        :param input_sizes: The input sizes of the model.
        :type input_sizes: Union[Dict[str, DimensionLike], SizeTypes]
        :param output_size: The output size of the model.
        :type output_size: Union[Dict[str, DimensionLike], SizeTypes]
        :param name: The name of the model.
        :type name: str
        :param checkpoint_folder: The folder where the checkpoints are saved.
        :type checkpoint_folder: str
        :param device: The device of the model. If None, the default device is used.
        :type device: torch.device
        :param input_transform: The transforms to apply to the inputs. The input_transform must work batch-wise.
        :type input_transform: Union[Dict[str, Callable], List[Callable]]
        :param output_transform: The transforms to apply to the outputs. The output_transform must work batch-wise.
        :type output_transform: Union[Dict[str, Callable], List[Callable]]

        :keyword kwargs: Additional arguments.
        """
        super(BaseModel, self).__init__(name=name)
        self._is_built = False
        self._given_input_transform = input_transform
        self._given_output_transform = output_transform
        self.input_transform: torch.nn.ModuleDict = None
        self.output_transform: torch.nn.ModuleDict = None
        self.input_sizes = input_sizes
        self.output_sizes = output_size
        self.name = name
        self.checkpoint_folder = checkpoint_folder
        self.kwargs = kwargs
        self._device = device
        if self._device is None:
            self._set_default_device_()
        self._to_device_transform = ToDevice(self.device)

    @property
    def input_sizes(self) -> Dict[str, int]:
        return self._input_sizes

    @input_sizes.setter
    def input_sizes(self, input_sizes: Union[Dict[str, DimensionLike], SizeTypes]):
        # if self.input_sizes is not None:
        # 	raise ValueError("Input sizes can only be set once.")
        if input_sizes is not None:
            self._input_sizes = self._format_sizes(input_sizes)

    @property
    def output_sizes(self) -> Dict[str, int]:
        return self._output_sizes

    @output_sizes.setter
    def output_sizes(self, output_size: Union[Dict[str, DimensionLike], SizeTypes]):
        if output_size is not None:
            self._output_sizes = self._format_sizes(output_size)

    @property
    def _ready(self):
        is_all_not_none = all([s is not None for s in [self._input_sizes, self._output_sizes]])
        if is_all_not_none:
            is_any_none = any([s is None for s in list(self._input_sizes.values()) + list(self._output_sizes.values())])
        else:
            is_any_none = True
        return is_all_not_none and not is_any_none

    @property
    def is_built(self) -> bool:
        return self._is_built

    @property
    def checkpoints_meta_path(self) -> str:
        """
        The path to the checkpoints meta file.

        :return: The path to the checkpoints meta file.
        :rtype: str
        """
        full_filename = (
            f"{self.name}{CheckpointManager.SUFFIX_SEP}{CheckpointManager.CHECKPOINTS_META_SUFFIX}"
        )
        return f"{self.checkpoint_folder}/{full_filename}.json"

    @property
    def device(self) -> torch.device:
        """
        :return: The device of the model.
        :rtype: torch.device
        """
        return self._device

    @device.setter
    def device(self, device: torch.device):
        """
        Set the device of the network.

        :param device: The device to set.
        :type device: torch.device

        :return: None
        """
        self.to(device, non_blocking=True)

    def to(self, device: torch.device, non_blocking: bool = True, *args, **kwargs):
        self._device = device
        for module in self.modules():
            if module is not self and getattr(module, "device", device).type != device.type:
                module.to(device, non_blocking=non_blocking)
        return super(BaseModel, self).to(device=device, non_blocking=non_blocking, *args, **kwargs)

    def _make_input_transform(
            self,
            input_transform: Union[Dict[str, Callable], List[Callable]]
    ) -> torch.nn.ModuleDict:
        """
        Make the input transform containing the transforms to apply to the inputs. If the input_transform is None,
        the default transform is used. If the input_transform is a list, it is converted to a dict. If the
        input_transform is a dict, it is copied. The keys of the input_transform are the names of the inputs if the
        network has inputs else the outputs if the network has outputs.

        :param input_transform: The input transform to use.
        :type input_transform: Union[Dict[str, Callable], List[Callable]]

        :return: The input transform.
        :rtype: torch.nn.ModuleDict
        """
        if len(self.input_sizes) > 0:
            transform_keys = list(self.input_sizes.keys())
        elif len(self.output_sizes) > 0:
            transform_keys = list(self.output_sizes.keys())
        else:
            transform_keys = []

        if input_transform is None:
            input_transform = self.get_default_input_transform()
        if isinstance(input_transform, list):
            default_transform = self.get_default_input_transform()
            if len(input_transform) < len(transform_keys):
                for i in range(len(input_transform), len(transform_keys)):
                    input_transform.append(default_transform[transform_keys[i]])
            input_transform = {in_name: t for in_name, t in zip(transform_keys, input_transform)}
        elif callable(input_transform):
            input_transform = {in_name: input_transform for in_name in transform_keys}
        if isinstance(input_transform, dict):
            assert all([in_name in input_transform for in_name in transform_keys]), \
                f"Input transform must contain all input names: {transform_keys}"
        else:
            raise TypeError(f"Input transform must be a dict or a list of callables. Got {type(input_transform)}.")

        for in_name, t in input_transform.items():
            if isinstance(t, torch.nn.Module):
                input_transform[in_name] = t.to(self.device)
            else:
                input_transform[in_name] = CallableToModuleWrapper(t).to(self.device)
        return torch.nn.ModuleDict(input_transform)

    def _make_output_transform(
            self,
            output_transform: Union[Dict[str, Callable], List[Callable]]
    ) -> torch.nn.ModuleDict:
        """
        Make the output transform containing the transforms to apply to the outputs. If the output_transform is None,
        the default transform is used. If the output_transform is a list, it is converted to a dict. If the
        output_transform is a dict, it is copied. The keys of the output_transform are the names of the outputs if the
        network has outputs.

        :param output_transform: The output transform to use.
        :type output_transform: Union[Dict[str, Callable], List[Callable]]

        :return: The output transform.
        :rtype: torch.nn.ModuleDict
        """
        if len(self.output_sizes) > 0:
            transform_keys = list(self.output_sizes.keys())
        else:
            transform_keys = []

        if output_transform is None:
            output_transform = self.get_default_output_transform()
        if isinstance(output_transform, list):
            default_transform = self.get_default_output_transform()
            if len(output_transform) < len(transform_keys):
                for i in range(len(output_transform), len(transform_keys)):
                    output_transform.append(default_transform[transform_keys[i]])
            output_transform = {in_name: t for in_name, t in zip(transform_keys, output_transform)}
        elif callable(output_transform):
            output_transform = {in_name: output_transform for in_name in transform_keys}
        if isinstance(output_transform, dict):
            assert all([in_name in output_transform for in_name in transform_keys]), \
                f"Output transform must contain all output names: {transform_keys}"
        else:
            raise TypeError(f"Output transform must be a dict or a list of callables. Got {type(output_transform)}.")

        for out_name, t in output_transform.items():
            if isinstance(t, torch.nn.Module):
                output_transform[out_name] = t.to(self.device)
            else:
                output_transform[out_name] = CallableToModuleWrapper(t).to(self.device)
        return torch.nn.ModuleDict(output_transform)

    def load_checkpoint(
            self,
            checkpoints_meta_path: Optional[str] = None,
            load_checkpoint_mode: LoadCheckpointMode = LoadCheckpointMode.BEST_ITR,
            verbose: bool = True
    ) -> dict:
        """
        Load the checkpoint from the checkpoints_meta_path. If the checkpoints_meta_path is None, the default
        checkpoints_meta_path is used.

        :param checkpoints_meta_path: The path to the checkpoints meta file.
        :type checkpoints_meta_path: Optional[str]
        :param load_checkpoint_mode: The mode to use when loading the checkpoint.
        :type load_checkpoint_mode: LoadCheckpointMode
        :param verbose: Whether to print the loaded checkpoint information.
        :type verbose: bool

        :return: The loaded checkpoint information.
        :rtype: dict
        """
        if checkpoints_meta_path is None:
            checkpoints_meta_path = self.checkpoints_meta_path
        with open(checkpoints_meta_path, "r+") as jsonFile:
            info: dict = json.load(jsonFile)
        save_name = CheckpointManager.get_save_name_from_checkpoints(info, load_checkpoint_mode)
        checkpoint_path = f"{self.checkpoint_folder}/{save_name}"
        if verbose:
            logging.info(f"Loading checkpoint from {checkpoint_path}")
        checkpoint = torch.load(checkpoint_path, map_location=self.device)
        self.load_state_dict(checkpoint[CheckpointManager.CHECKPOINT_STATE_DICT_KEY], strict=True)
        return checkpoint

    def get_default_input_transform(self) -> Dict[str, nn.Module]:
        """
        Get the default input transform. The default input transform is a to tensor transform.

        :return: The default input transform.
        :rtype: Dict[str, nn.Module]
        """
        if len(self.input_sizes) > 0:
            transform_keys = list(self.input_sizes.keys())
        elif len(self.output_sizes) > 0:
            transform_keys = list(self.output_sizes.keys())
        else:
            transform_keys = []
        return {
            in_name: Compose([
                ToTensor(dtype=torch.float32),
            ])
            for in_name in transform_keys
        }

    def get_default_output_transform(self) -> Dict[str, nn.Module]:
        """
        Get the default output transform. The default output transform is an identity transform.

        :return: The default output transform.
        :rtype: Dict[str, nn.Module]
        """
        if len(self.output_sizes) > 0:
            transform_keys = list(self.output_sizes.keys())
        else:
            transform_keys = []
        return {
            in_name: IdentityTransform()
            for in_name in transform_keys
        }

    def apply_input_transform(self, inputs: Dict[str, Any]) -> Dict[str, torch.Tensor]:
        """
        Apply the input transform to the inputs.

        :param inputs: dict of inputs of shape (batch_size, *input_size)
        :type inputs: Dict[str, Any]

        :return: The input of the network with the same shape as the input.
        :rtype: Dict[str, torch.Tensor]
        """
        assert all([in_name in self.input_transform for in_name in inputs]), \
            f"Inputs must be all in input names: {self.input_transform.keys()}"
        inputs = {
            in_name: self.input_transform[in_name](in_batch)
            for in_name, in_batch in inputs.items()
        }
        return inputs

    def apply_output_transform(self, outputs: Dict[str, Any]) -> Dict[str, torch.Tensor]:
        """
        Apply the output transform to the outputs.

        :param outputs: dict of outputs of shape (batch_size, *output_size).
        :type outputs: Dict[str, Any]

        :return: The output of the network transformed.
        :rtype: Dict[str, torch.Tensor]
        """
        assert all([out_name in self.output_transform for out_name in outputs]), \
            f"Outputs must be all in output names: {self.output_transform.keys()}"
        outputs = {
            out_name: self.output_transform[out_name](out_batch)
            for out_name, out_batch in outputs.items()
        }
        return outputs

    def _add_to_device_transform_(self):
        """
        Add the to_device_transform to the input transforms.

        :return: None
        """
        for in_name, trans in self.input_transform.items():
            list_of_transforms = ravel_compose_transforms(self.input_transform[in_name])
            list_of_transforms.append(self._to_device_transform)
            self.input_transform[in_name] = list_of_callable_to_sequential(list_of_transforms)
            trans.to(self.device)

    def _remove_to_device_transform_(self):
        """
        Remove the to_device transform from the transforms.

        :return: None
        """
        for in_name, trans in self.input_transform.items():
            if self._to_device_transform:
                list_of_transforms = ravel_compose_transforms(self.input_transform[in_name])
                list_of_transforms.remove(self._to_device_transform)
                self.input_transform[in_name] = list_of_callable_to_sequential(list_of_transforms)

    def _set_default_device_(self):
        """
        Set the default device of the network. The default device will be cuda if available and cpu otherwise.

        :return: None
        """
        self._device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

    def infer_sizes_from_inputs(self, inputs: Union[Dict[str, Any], torch.Tensor]):
        """
        Infer the input and output sizes from the inputs.

        :param inputs: The inputs of the network.
        :type inputs: Union[Dict[str, Any], torch.Tensor]

        :return: None
        """
        if isinstance(inputs, torch.Tensor):
            inputs = {
                "0": inputs
            }
        self.input_sizes = {k: v.shape[1:] for k, v in inputs.items()}

    def build(self, *args, **kwargs) -> 'BaseModel':
        """
        Build the network.

        :param args: Not used.
        :param kwargs: Not used.

        :return: The network.
        :rtype: BaseModel
        """
        self._is_built = True
        self.input_transform: Dict[str, Callable] = self._make_input_transform(self._given_input_transform)
        self.output_transform: Dict[str, Callable] = self._make_output_transform(self._given_output_transform)
        self._add_to_device_transform_()
        self.device = self._device
        return self

    def __call__(self, inputs: Union[Dict[str, Any], torch.Tensor], *args, **kwargs):
        if not self._is_built:
            self.infer_sizes_from_inputs(inputs)
            self.build()
        return super(BaseModel, self).__call__(inputs, *args, **kwargs)

    def forward(self, inputs: Union[Dict[str, Any], torch.Tensor], **kwargs) -> Dict[str, torch.Tensor]:
        raise NotImplementedError()

    def get_prediction_trace(
            self,
            inputs: Union[Dict[str, Any], torch.Tensor],
            **kwargs
    ) -> Union[Dict[str, torch.Tensor], torch.Tensor]:
        """
        Get the prediction trace of the network.

        :param inputs: The inputs of the network.
        :type inputs: Union[Dict[str, Any], torch.Tensor]
        :param kwargs: Additional arguments.

        :return: The prediction trace.
        :rtype: Union[Dict[str, torch.Tensor], torch.Tensor]
        """
        raise NotImplementedError()

    def get_raw_prediction(
            self,
            inputs: torch.Tensor,
            re_outputs_trace: bool = True,
            re_hidden_states: bool = True
    ) -> Union[Tuple[Any, Any, Any], Tuple[Any, Any], Any]:
        """
        The raw prediction of the network.

        :param inputs: The inputs of the network.
        :type inputs: torch.Tensor
        :param re_outputs_trace: If True, the outputs trace will be returned.
        :type re_outputs_trace: bool
        :param re_hidden_states: If True, the hidden states will be returned.
        :type re_hidden_states: bool

        :return: The raw prediction.
        :rtype: Union[Tuple[Any, Any, Any], Tuple[Any, Any], Any]
        """
        raise NotImplementedError()

    def get_prediction_proba(
            self,
            inputs: torch.Tensor,
            re_outputs_trace: bool = True,
            re_hidden_states: bool = True
    ) -> Union[Tuple[Any, Any, Any], Tuple[Any, Any], Any]:
        """
        Get the prediction probabilities of the network.

        :param inputs: The inputs of the network.
        :type inputs: torch.Tensor
        :param re_outputs_trace: If True, the outputs trace will be returned.
        :type re_outputs_trace: bool
        :param re_hidden_states: If True, the hidden states will be returned.
        :type re_hidden_states: bool

        :return: The prediction probabilities.
        :rtype: Union[Tuple[Any, Any, Any], Tuple[Any, Any], Any]
        """
        m, *outs = self.get_raw_prediction(inputs, re_outputs_trace, re_hidden_states)
        if isinstance(m, torch.Tensor):
            proba = torch.softmax(m, dim=-1)
        elif isinstance(m, dict):
            proba = {
                k: torch.softmax(v, dim=-1)
                for k, v in m.items()
            }
        else:
            raise ValueError("m must be a torch.Tensor or a dictionary")
        if re_outputs_trace or re_hidden_states:
            return proba, *outs
        return proba

    def get_prediction_log_proba(
            self,
            inputs: torch.Tensor,
            re_outputs_trace: bool = True,
            re_hidden_states: bool = True
    ) -> Union[Tuple[Any, Any, Any], Tuple[Any, Any], Any]:
        """
        Get the prediction log probabilities of the network.

        :param inputs: The inputs of the network.
        :type inputs: torch.Tensor
        :param re_outputs_trace: If True, the outputs trace will be returned.
        :type re_outputs_trace: bool
        :param re_hidden_states: If True, the hidden states will be returned.
        :type re_hidden_states: bool

        :return: The prediction log probabilities.
        :rtype: Union[Tuple[Any, Any, Any], Tuple[Any, Any], Any]
        """
        m, *outs = self.get_raw_prediction(inputs, re_outputs_trace, re_hidden_states)
        if isinstance(m, torch.Tensor):
            log_proba = F.log_softmax(m, dim=-1)
        elif isinstance(m, dict):
            log_proba = {
                k: F.log_softmax(v, dim=-1)
                for k, v in m.items()
            }
        else:
            raise ValueError("m must be a torch.Tensor or a dictionary")
        if re_outputs_trace or re_hidden_states:
            return log_proba, *outs
        return log_proba

    def soft_update(self, other: 'BaseModel', tau: float = 1e-2) -> None:
        """
        Copies the weights from the other network to this network with a factor of tau.

        :param other: The other network.
        :type other: 'BaseModel'
        :param tau: The factor of the copy.
        :type tau: float

        :return: None
        """
        with torch.no_grad():
            for param, other_param in zip(self.parameters(), other.parameters()):
                param.data.copy_((1 - tau) * param.data + tau * other_param.data)

    def hard_update(self, other: 'BaseModel') -> None:
        """
        Copies the weights from the other network to this network.

        :param other: The other network.
        :type other: 'BaseModel'

        :return: None
        """
        with torch.no_grad():
            self.load_state_dict(other.state_dict())

    def to_onnx(self, in_viz=None):
        """
        Creates an ONNX model from the network.

        :param in_viz: The input to visualize.
        :type in_viz: Any

        :return: The ONNX model.
        """
        if in_viz is None:
            in_viz = torch.randn((1, self.input_sizes), device=self._device)
        torch.onnx.export(
            self,
            in_viz,
            f"{self.checkpoint_folder}/{self.name}.onnx",
            verbose=True,
            input_names=None,
            output_names=None,
            opset_version=11
        )



from typing import Callable, Dict, List, Optional, Union

import torch

from . import Trainer
from ..learning_algorithms.bptt import BPTT
from ..learning_algorithms.learning_algorithm import LearningAlgorithm
from ..transforms.base import ToTensor
from ..metrics import ClassificationMetrics


class ClassificationTrainer(Trainer):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("y_transform", ToTensor(dtype=torch.long))
        kwargs.setdefault("predict_method", "get_prediction_log_proba")
        super().__init__(*args, **kwargs)

    def _set_default_criterion(self, criterion: Optional[torch.nn.Module]) -> torch.nn.Module:
        if criterion is None:
            if isinstance(self.model.output_sizes, dict):
                criterion = {
                    k: torch.nn.NLLLoss() for k in self.model.output_sizes
                }
            elif isinstance(self.model.output_sizes, int):
                criterion = torch.nn.NLLLoss()
            else:
                raise ValueError("Unknown criterion type")
        return criterion

    def _set_default_metrics(self, metrics: Optional[List[Callable]]):
        if metrics is None:
            metrics = [ClassificationMetrics(self.model)]
        return metrics

    def _maybe_add_learning_algorithm(self, learning_algorithm: Optional[LearningAlgorithm]) -> None:
        if len(self.learning_algorithms) == 0 and learning_algorithm is None:
            learning_algorithm = BPTT(
                optimizer=torch.optim.Adam(self.model.parameters()),
                criterion=self._set_default_criterion(None)
            )
        if learning_algorithm is not None:
            self.callbacks.append(learning_algorithm)






from typing import Callable, Optional

from .base_callback import BaseCallback


class EventOnMetricThreshold(BaseCallback):
    r"""
    This class is a callback that call an event if the metric value reach a given threshold at the end of each
    iteration. Note that the trainer will be passed as the first argument to the event. In addition, if the
    given metric is not present in the training state, the event will not be called. Finally, the output of the event
    will be pass to the :meth:`on_pbar_update` method.

    :Attributes:
        - **metric_name** (str): The name of the metric to monitor.
        - **threshold** (float): The threshold value.
        - **event** (Callable): The event to call.
        - **event_args** (tuple): The arguments to pass to the event.
        - **event_kwargs** (dict): The keyword arguments to pass to the event.
    """
    def __init__(
            self,
            metric_name: str,
            threshold: float,
            *,
            event: Callable,
            event_args: Optional[tuple] = None,
            event_kwargs: Optional[dict] = None,
            minimize_metric: bool = True,
            do_once: bool = False,
            **kwargs
    ):
        """
        Constructor for the EventOnMetricThreshold class.

        :param metric_name: The name of the metric to monitor.
        :type metric_name: str
        :param threshold: The threshold value.
        :type threshold: float
        :param event: The event to call.
        :type event: Callable
        :param event_args: The arguments to pass to the event.
        :type event_args: tuple
        :param event_kwargs: The keyword arguments to pass to the event.
        :type event_kwargs: dict
        :param kwargs: The keyword arguments to pass to the base class.
        """
        super().__init__(**kwargs)
        self.metric_name = metric_name
        self.threshold = threshold
        self.event = event
        self.event_args = event_args if event_args is not None else tuple()
        self.event_kwargs = event_kwargs if event_kwargs is not None else dict()
        self._current_event_output = None
        self.minimize_metric = minimize_metric
        self.do_once = do_once
        self._has_triggered = False

    def on_iteration_end(self, trainer, **kwargs):
        """
        Check if the metric value reach the threshold.

        :param trainer: The trainer object.
        :type trainer: Trainer

        :return: None
        """
        if self.do_once and self._has_triggered:
            return
        if self.metric_name in trainer.current_training_state.itr_metrics:
            metric_value = trainer.current_training_state.itr_metrics[self.metric_name]
            if self.minimize_metric:
                threshold_reached = metric_value <= self.threshold
            else:
                threshold_reached = metric_value >= self.threshold
            if threshold_reached:
                self._current_event_output = self.event(trainer, *self.event_args, **self.event_kwargs)
                self._has_triggered = True

    def on_pbar_update(self, trainer, **kwargs) -> dict:
        """
        Update the progress bar with the current learning algorithm.

        :param trainer: The trainer object.
        :type trainer: Trainer

        :return: The progress bar update.
        :rtype: dict
        """
        if self._current_event_output is not None:
            return {self.name: self._current_event_output}
        return dict()







"""
NeuroTorch: A Python library for machine learning and neuroscience.
"""

__author__ = "Jérémie Gince"
__email__ = "gincejeremie@gmail.com"
__copyright__ = "Copyright 2022, Jérémie Gince"
__license__ = "Apache 2.0"
__url__ = "https://github.com/NeuroTorch/NeuroTorch"
__version__ = "0.0.1-beta5"


from .dimension import (
    Dimension,
    DimensionProperty,
    Size,
    DimensionLike,
    DimensionsLike,
)

from .modules.sequential import (
    Sequential,
)

from .modules.sequential_rnn import (
    SequentialRNN,
)

from .modules.layers import (
    LayerType,
    Linear,
    LILayer,
    LIFLayer,
    SpyLILayer,
    SpyLIFLayer,
    SpyALIFLayer,
    ALIFLayer,
    SpyLIFLayerLPF,
    SpyALIFLayerLPF,
    LIFLayerLPF,
    ALIFLayerLPF,
    WilsonCowanLayer,
    WilsonCowanCURBDLayer,
)

from .regularization import (
    RegularizationList,
    L1,
    L2,
)

from .regularization.connectome import (
    DaleLaw,
    DaleLawL2,
)

from .trainers import (
    Trainer,
    ClassificationTrainer,
    RegressionTrainer,
    TrainingState,
)

from .transforms import (
    to_tensor,
    to_numpy,
    IdentityTransform,
    ToDevice,
    ToTensor,
)


from .metrics import (
    losses,
)

from .callbacks import (
    TrainingHistory,
    LoadCheckpointMode,
    CheckpointManager,
)

from . import init

from .learning_algorithms import (
    BPTT,
    TBPTT,
    Eprop,
    RLS,
)

from .visualisation import (
    Visualise,
    VisualiseKMeans,
    VisualisePCA,
    VisualiseUMAP,
)

from . import utils

from .utils import (
    set_seed,
)

import warnings

warnings.filterwarnings("ignore", category=Warning, module="docutils")
warnings.filterwarnings("ignore", category=Warning, module="sphinx")



import enum
from typing import Optional

from .base import (
    BaseLayer,
    BaseNeuronsLayer,
)

from .classical import (
    Linear,
    LinearRNN,
)

from .spiking import (
    LIFLayer,
    ALIFLayer,
    SpyLIFLayer,
    SpyALIFLayer,
    IzhikevichLayer,
)

from .spiking_lpf import (
    SpyLIFLayerLPF,
    SpyALIFLayerLPF,
    LIFLayerLPF,
    ALIFLayerLPF,
)

from .wilson_cowan import (
    WilsonCowanLayer,
    WilsonCowanCURBDLayer,
)

from .leaky_integrate import (
    LILayer,
    SpyLILayer,
)


class LayerType(enum.Enum):
    LIF = 0
    ALIF = 1
    Izhikevich = 2
    LI = 3
    SpyLIF = 4
    SpyLI = 5
    SpyALIF = 6

    @classmethod
    def from_str(cls, name: str) -> Optional['LayerType']:
        """
        Get the LayerType from a string.

        :param name: The name of the LayerType.
        :type name: str

        :return: The LayerType.
        :rtype: Optional[LayerType]
        """
        if isinstance(name, LayerType):
            return name
        if name.startswith(cls.__name__):
            name = name.removeprefix(f"{cls.__name__}.")
        if name not in cls.__members__:
            return None
        return cls[name]


LayerType2Layer = {
    LayerType.LIF: LIFLayer,
    LayerType.ALIF: ALIFLayer,
    LayerType.Izhikevich: IzhikevichLayer,
    LayerType.LI: LILayer,
    LayerType.SpyLIF: SpyLIFLayer,
    LayerType.SpyALIF: SpyALIFLayer,
    LayerType.SpyLI: SpyLILayer,
}



import warnings
from collections import OrderedDict
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Tuple, Type, Union
from typing import OrderedDict as OrderedDictType

import torch
import torch.nn.functional as F
from torch import Tensor, nn

from . import (
    BaseLayer,
    BaseModel,
    LayerType,
    LayerType2Layer,
    SpikeFuncType,
    SpikeFunction
)
from .base import NamedModule
from .wrappers import NamedModuleWrapper
from ..dimension import Dimension


Acceptable_Spike_Func = Union[Type[SpikeFunction], SpikeFuncType]
Acceptable_Spike_Funcs = Union[Acceptable_Spike_Func, Iterable[Acceptable_Spike_Func]]
Acceptable_Layer_Type = Union[Type[BaseLayer], LayerType]
Acceptable_Layer_Types = Union[Acceptable_Layer_Type, Iterable[Acceptable_Layer_Type]]
IntDimension = Union[int, Dimension]


class Sequential(BaseModel):
    """
    The Sequential is a neural network that is constructed by stacking layers.

    .. image:: ../../images/modules/Sequential_model_schm.png
        :width: 300
        :align: center

    :Attributes:
        - :attr:`input_layers` (torch.nn.ModuleDict): The input layers of the model.
        - :attr:`hidden_layers` (torch.nn.ModuleList): The hidden layers of the model.
        - :attr:`output_layers` (torch.nn.ModuleDict): The output layers of the model.
        - :attr:`foresight_time_steps` (int): The number of time steps that the model will forecast.
    """

    @staticmethod
    def _format_input_output_layers(
            layers: Iterable[Union[Iterable[torch.nn.Module], torch.nn.Module]],
            default_prefix_layer_name: str = "layer",
    ) -> OrderedDictType[str, NamedModule]:
        """
        Format the input or output layers. The format is an ordered dictionary of the form {layer_name: layer}.

        :param layers: The input or output layers.
        :type layers: Iterable[Union[Iterable[torch.nn.Module], torch.nn.Module]]
        :param default_prefix_layer_name: The default prefix of the layer name. The prefix is used when the name of
        the layer is not specified.
        :type default_prefix_layer_name: str

        :return: The formatted input or output layers.
        :rtype: OrderedDict[str, NamedModule]
        """
        layers: Iterable[torch.nn.Module] = [layers] if not isinstance(layers, (Iterable, Mapping)) else layers
        if isinstance(layers, Mapping):
            layers: OrderedDictType[str, NamedModule] = OrderedDict(
                (k, (v if isinstance(v, NamedModule) else NamedModuleWrapper(v))) for k, v in layers.items()
            )
            for layer_key, layer in layers.items():
                if not layer.name_is_set:
                    layer.name = layer_key
            assert all(layer_key == layer.name for layer_key, layer in layers.items()), \
                "The layer names must be the same as the keys."
        else:
            layers: Iterable[NamedModule] = [
                (layer if isinstance(layer, NamedModule) else NamedModuleWrapper(layer)) for layer in layers
            ]
            for layer_idx, layer in enumerate(layers):
                if not layer.name_is_set:
                    layer.name = f"{default_prefix_layer_name}_{layer_idx}"
            assert len([layer.name for layer in layers]) == len(set([layer.name for layer in layers])), \
                "There are layers with the same name. Please specify the names of the layers without duplicates."
            layers: OrderedDict[str, NamedModule] = OrderedDict((layer.name, layer) for layer in layers)
        return layers

    @staticmethod
    def _format_hidden_layers(
            layers: Iterable[torch.nn.Module],
            default_prefix_layer_name: str = "hidden",
    ) -> List[NamedModule]:
        """
        Format the hidden layers. The format is a list of the form [layer, ...].

        :param layers: The hidden layers.
        :type layers: Iterable[torch.nn.Module]
        :param default_prefix_layer_name: The default prefix of the layer name. The prefix is used when the name of
        the layer is not specified.
        :type default_prefix_layer_name: str

        :return: The formatted hidden layers.
        :rtype: List[NamedModule]
        """
        layers: Iterable[NamedModule] = [
            (layer if isinstance(layer, NamedModule) else NamedModuleWrapper(layer)) for layer in layers
        ]
        for i, layer in enumerate(layers):
            if not layer.name_is_set:
                layer.name = f"{default_prefix_layer_name}_{i}"
        return list(layers)

    @staticmethod
    def _format_layers(
            layers: Iterable[Union[Iterable[torch.nn.Module], torch.nn.Module]]
    ) -> Tuple[OrderedDict, List, OrderedDict]:
        """
        Format the given layers. The format is a tuple of the form:

        ::

            (
                OrderedDict({input_layer_name: input_layer}),
                List(hidden_layers),
                OrderedDict({output_layer_name: output_layer}),
            )

        :param layers: The layers.
        :type layers: Iterable[Union[Iterable[BaseLayer], BaseLayer]]

        :return: The formatted layers.
        :rtype: Tuple[OrderedDict, List, OrderedDict]
        """
        if not isinstance(layers, Iterable):
            layers = [layers]
        if len(layers) > 1:
            input_layers = Sequential._format_input_output_layers(layers[0], "input")
        else:
            input_layers = nn.ModuleDict()

        if len(layers) > 2:
            hidden_layers = layers[1:-1]
            if not isinstance(hidden_layers, Iterable):
                hidden_layers = [hidden_layers]
        else:
            hidden_layers = []
        hidden_layers = Sequential._format_hidden_layers(hidden_layers)

        output_layers = Sequential._format_input_output_layers(layers[-1], "output")
        return input_layers, hidden_layers, output_layers

    @staticmethod
    def _layers_containers_to_modules(
            inputs_layers: OrderedDict,
            hidden_layers: List,
            outputs_layers: OrderedDict
    ) -> Tuple[nn.ModuleDict, nn.ModuleList, nn.ModuleDict]:
        """
        Convert the input, hidden and output layers containers to modules.

        :param inputs_layers: The input layers.
        :type inputs_layers: OrderedDict
        :param hidden_layers: The hidden layers.
        :type hidden_layers: List
        :param outputs_layers: The output layers.
        :type outputs_layers: OrderedDict

        :return: The input, hidden and output layers modules.
        :rtype: Tuple[nn.ModuleDict, nn.ModuleList, nn.ModuleDict]
        """
        input_layers = nn.ModuleDict(inputs_layers)
        hidden_layers = nn.ModuleList(hidden_layers)
        output_layers = nn.ModuleDict(outputs_layers)
        return input_layers, hidden_layers, output_layers

    @staticmethod
    def _format_layer_type_(
            layer_type: Optional[Acceptable_Layer_Type]
    ) -> Optional[Type[BaseLayer]]:
        warnings.warn(
            "This function is not used anymore in the SequentialModel.",
            DeprecationWarning
        )
        if isinstance(layer_type, LayerType):
            layer_type = LayerType2Layer[layer_type]
        return layer_type

    @staticmethod
    def _format_hidden_neurons_(n_hidden_neurons: Optional[Union[int, Iterable[int]]]) -> List[int]:
        warnings.warn(
            "This function is not used anymore in the SequentialModel.",
            DeprecationWarning
        )
        if n_hidden_neurons is None:
            return []
        if not isinstance(n_hidden_neurons, Iterable):
            n_hidden_neurons = [n_hidden_neurons]
        return n_hidden_neurons

    def __init__(
            self,
            layers: Iterable[Union[Iterable[torch.nn.Module], torch.nn.Module]],
            name: str = "Sequential",
            checkpoint_folder: str = "checkpoints",
            device: Optional[torch.device] = None,
            input_transform: Optional[Union[Dict[str, Callable], List[Callable]]] = None,
            output_transform: Optional[Union[Dict[str, Callable], List[Callable]]] = None,
            **kwargs
    ):
        """
        The Sequential is a neural network that is constructed by stacking layers.

        :param layers: The layers to be used in the model. One of the following structure is expected:
        ::

            layers = [
                [*inputs_layers, ],
                *hidden_layers,
                [*output_layers, ]
            ]
            or
            layers = [
                input_layer,
                *hidden_layers,
                output_layer
            ]

        :type layers: Iterable[Union[Iterable[torch.nn.Module], torch.nn.Module]]
        :param name: The name of the model.
        :type name: str
        :param checkpoint_folder: The folder where the checkpoints are saved.
        :type checkpoint_folder: str
        :param device: The device to use.
        :type device: torch.device
        :param input_transform: The transform to apply to the input. The input_transform must work on a single datum.
        :type input_transform: Union[Dict[str, Callable], List[Callable]]
        :param output_transform: The transform to apply to the output trace. The output_transform must work batch-wise.
        :type output_transform: Union[Dict[str, Callable], List[Callable]]
        :param kwargs: Additional keyword arguments.
        """
        input_layers, hidden_layers, output_layers = self._format_layers(layers)
        self._ordered_inputs_names = [layer.name for _, layer in input_layers.items()]
        self._ordered_outputs_names = [layer.name for _, layer in output_layers.items()]
        super(Sequential, self).__init__(
            # TODO: automatically find the first layer in the network et get its input size.
            input_sizes={layer.name: getattr(layer, "input_size", None) for _, layer in input_layers.items()},
            output_size={layer.name: getattr(layer, "output_size", None) for _, layer in output_layers.items()},
            name=name,
            checkpoint_folder=checkpoint_folder,
            device=device,
            input_transform=input_transform,
            output_transform=output_transform,
            **kwargs
        )
        self._default_n_hidden_neurons = self.kwargs.get("n_hidden_neurons", 128)
        self.input_layers, self.hidden_layers, self.output_layers = self._layers_containers_to_modules(
            input_layers, hidden_layers, output_layers
        )
        assert len(self.get_all_layers_names()) == len(set(self.get_all_layers_names())), \
            "There are layers with the same name."

    @BaseModel.device.setter
    def device(self, device: torch.device):
        """
        Set the device of the model and all its layers.

        :param device: The device to use.
        :type device: torch.device

        :return: None
        """
        BaseModel.device.fset(self, device)
        for layer in self.get_all_layers():
            layer.device = device

    def get_all_layers(self) -> List[nn.Module]:
        """
        Get all the layers of the model as a list. The order of the layers is the same as the order of the layers in the
        model.

        :return: A list of all the layers of the model.
        :rtype: List[nn.Module]
        """
        return list(self.input_layers.values()) + list(self.hidden_layers) + list(self.output_layers.values())

    def get_layers(self, layer_names: Optional[List[str]] = None) -> List[nn.Module]:
        """
        Get the layers with the specified names.

        :param layer_names: The names of the layers to get.
        :type layer_names: Optional[List[str]]

        :return: The layers with the specified names.
        :rtype: List[nn.Module]
        """
        if layer_names is None:
            return self.get_all_layers()
        return [self.get_layer(layer_name) for layer_name in layer_names]

    def get_all_layers_names(self) -> List[str]:
        """
        Get all the names of the layers of the model. The order of the layers is the same as the order of the layers in
        the model.

        :return: A list of all the names of the layers of the model.
        :rtype: List[str]
        """
        return [layer.name for layer in self.get_all_layers()]

    def get_dict_of_layers(self) -> Dict[str, nn.Module]:
        """
        Get all the layers of the model as a dictionary. The order of the layers is the same as the order of the layers
        in the model. The keys of the dictionary are the names of the layers.

        :return: A dictionary of all the layers of the model.
        :rtype: Dict[str, nn.Module]
        """
        return {layer.name: layer for layer in self.get_all_layers()}

    def get_layer(self, name: Optional[str] = None) -> nn.Module:
        """
        Get a layer of the model. If the name is None, the first layer is returned which is useful when the model has
        only one layer.

        :param name: The name of the layer.
        :type name: str

        :return: The layer with the given name. If the name is None, the first layer is returned.
        :rtype: nn.Module
        """
        if name is None:
            return self.get_all_layers()[0]
        else:
            return self.get_dict_of_layers()[name]

    def __getitem__(self, name: Optional[str]) -> nn.Module:
        """
        Get a layer of the model. If the name is None, the first layer is returned which is useful when the model has
        only one layer.

        :param name: The name of the layer.
        :type name: str

        :return: The layer with the given name. If the name is None, the first layer is returned.
        :rtype: nn.Module
        """
        return self.get_layer(name)

    def infer_sizes_from_inputs(self, inputs: Union[Dict[str, Any], torch.Tensor]):
        """
        Infer the sizes of the inputs layers from the inputs of the network. The sizes of the inputs layers are set to
        the size of the inputs without the batch dimension.

        :param inputs: The inputs of the network.
        :type inputs: Union[Dict[str, Any], torch.Tensor]

        :return: None
        """
        if isinstance(inputs, torch.Tensor):
            inputs = {
                layer_name: inputs
                for layer_name, _ in self.input_layers.items()
            }
        self.input_sizes = {k: v.shape[1:] for k, v in inputs.items()}

    def initialize_weights_(self):
        """
        Initialize the weights of the layers of the model.

        :return: None
        """
        for layer in self.get_all_layers():
            if hasattr(layer, "initialize_weights_") and callable(layer.initialize_weights_):
                layer.initialize_weights_()

    def _format_single_inputs(self, inputs: torch.Tensor, **kwargs) -> torch.Tensor:
        """
        Cast the inputs to float tensor.

        :param inputs: Inputs tensor.
        :type inputs: torch.Tensor

        :return: Formatted Input tensor.
        :rtype: torch.Tensor
        """
        return inputs.float()

    def _format_inputs(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        Return the formatted inputs formatted by self._format_single_inputs.

        :param inputs: Inputs dictionary.
        :type inputs: Dict[str, torch.Tensor]

        :return: Formatted inputs dictionary.
        :rtype: Dict[str, torch.Tensor]
        """
        return {k: self._format_single_inputs(in_tensor) for k, in_tensor in inputs.items()}

    def _inputs_to_dict(self, inputs: Union[Dict[str, Any], torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        Transform the inputs tensor into dictionary of tensors.

        :param inputs: The inputs of the network.
        :type inputs: Union[Dict[str, Any], torch.Tensor]

        :return: The transformed inputs.
        :rtype: Dict[str, torch.Tensor]
        """
        keys = list(self.input_layers.keys())
        if len(keys) == 0 and len(list(self.output_layers.keys())) > 0:
            keys = self.output_layers.keys()
        if isinstance(inputs, torch.Tensor):
            inputs = {k: inputs for k in keys}
        else:
            if set(inputs.keys()) != set(keys):
                raise ValueError("inputs must have the same keys as the input layers")
        return inputs

    def build_layers(self):
        """
        Build the layers of the model.

        :return: None
        """
        for layer in self.get_all_layers():
            if hasattr(layer, "build"):
                if callable(layer.build):
                    if not getattr(layer, "is_built", False):
                        layer.build()

    def build(self) -> 'Sequential':
        """
        Build the network and all its layers.

        :return: The network.
        :rtype: Sequential
        """
        super(Sequential, self).build()
        self._infer_and_set_sizes_of_all_layers()
        self.build_layers()
        self.initialize_weights_()
        self.device = self._device
        return self

    def _infer_and_set_sizes_of_all_layers(self):
        """
        Infer the sizes of all layers and set them.

        :return: None
        """
        inputs_layers_out_sum = 0
        inputs_sum_valid = True
        for layer_name, layer in self.input_layers.items():
            if hasattr(layer, "input_size") and layer.input_size is None:
                layer.input_size = self.input_sizes[layer_name]
            if hasattr(layer, "output_size"):
                if layer.output_size is None:
                    layer.output_size = self._default_n_hidden_neurons
                inputs_layers_out_sum += int(layer.output_size)
            else:
                inputs_sum_valid = False

        last_hidden_out_size = inputs_layers_out_sum
        for layer_idx, layer in enumerate(self.hidden_layers):
            if layer_idx == 0:
                if hasattr(layer, "input_size") and layer.input_size is None and inputs_sum_valid:
                    layer.input_size = last_hidden_out_size
            # layer.input_size = inputs_layers_out_sum
            elif (
                    hasattr(self.hidden_layers[layer_idx - 1], "output_size")
                    and hasattr(layer, "input_size")
                    and layer.input_size is None
            ):
                layer.input_size = self.hidden_layers[layer_idx - 1].output_size
            if hasattr(layer, "output_size") and layer.output_size is None:
                layer.output_size = self._default_n_hidden_neurons
            if hasattr(layer, "output_size"):
                last_hidden_out_size = int(layer.output_size)

        for layer_name, layer in self.output_layers.items():
            if hasattr(layer, "input_size") and layer.input_size is None:
                layer.input_size = last_hidden_out_size
            if hasattr(layer, "output_size") and layer.output_size is None:
                if self.output_sizes is None or self.output_sizes[layer_name] is None:
                    warnings.warn(
                        f"output_size is not set for layer {layer_name}. It will be set to {last_hidden_out_size}"
                    )
                    layer.output_size = last_hidden_out_size
                else:
                    layer.output_size = self.output_sizes[layer_name]
            if self.output_sizes is None:
                self.output_sizes = {layer_name: layer.output_size} if hasattr(layer, "output_size") else {}
            elif hasattr(layer, "output_size"):
                self.output_sizes[layer_name] = layer.output_size

    def forward(
            self,
            inputs: Union[Dict[str, Any], torch.Tensor],
            **kwargs
    ) -> Dict[str, torch.Tensor]:
        """
        Forward pass of the model.

        :param inputs: The inputs to the model where the dimensions are {input_name: (batch_size, input_size)}.
        :type inputs: Union[Dict[str, Any], torch.Tensor]
        :param kwargs: Additional arguments for the forward pass.

        :return: A dictionary of outputs where the values are the names of the layers and the values are the outputs
                of the layers.
        :rtype: Dict[str, torch.Tensor]
        """
        inputs = self._inputs_to_dict(inputs)
        inputs = self.apply_input_transform(inputs)
        inputs = self._format_inputs(inputs)
        outputs: Dict[str, torch.Tensor] = {}

        features_list = []
        for layer_name, layer in self.input_layers.items():
            features = layer(inputs[layer_name])
            features_list.append(features)
        if features_list:
            forward_tensor = torch.concat(features_list, dim=-1)
        else:
            forward_tensor = torch.concat([inputs[in_name] for in_name in inputs], dim=-1)

        for layer_idx, layer in enumerate(self.hidden_layers):
            forward_tensor = layer(forward_tensor)

        for layer_name, layer in self.output_layers.items():
            out = layer(forward_tensor)
            outputs[layer_name] = out

        outputs_tensor = self.apply_output_transform(outputs)
        return outputs_tensor

    def get_raw_prediction(
            self,
            inputs: torch.Tensor,
            **kwargs
    ) -> Any:
        """
        Get the raw prediction of the model which is the output of the forward pass.

        :param inputs: inputs to the network.
        :type inputs: torch.Tensor

        :return: the raw prediction of the model.
        :rtype: Any
        """
        outputs = self(inputs.to(self._device))
        return outputs

    def get_prediction_proba(
            self,
            inputs: torch.Tensor,
            **kwargs
    ) -> Any:
        """
        Get the prediction probability of the model which is the softmax of the output of the forward pass.
        The softmax is performed on the last dimension. This method is generally used for classification.

        :param inputs: inputs to the network.
        :type inputs: torch.Tensor

        :return: the prediction probability of the model.
        :rtype: Union[Tuple[Any, Any, Any], Tuple[Any, Any], Any]
        """
        outs = self.get_raw_prediction(inputs, **kwargs)
        if isinstance(outs, (list, tuple)):
            m = outs[0]
        else:
            m = outs
        if isinstance(m, torch.Tensor):
            proba = torch.softmax(m, dim=-1)
        elif isinstance(m, dict):
            proba = {
                k: torch.softmax(v, dim=-1)
                for k, v in m.items()
            }
        else:
            raise ValueError("m must be a torch.Tensor or a dictionary")
        return proba

    def get_prediction_log_proba(
            self,
            inputs: torch.Tensor,
            **kwargs
    ) -> Union[Tuple[Tensor, Any, Any], Tuple[Tensor, Any], Tensor]:
        """
        Get the prediction log probability of the model which is the log softmax of the output of the forward pass.
        The log softmax is performed on the last dimension. This method is generally used for training in classification
        task.

        :param inputs: inputs to the network.
        :type inputs: torch.Tensor

        :return: the prediction log probability of the model.
        :rtype: Union[tuple[Tensor, Any, Any], tuple[Tensor, Any], Tensor]
        """
        outs = self.get_raw_prediction(inputs, **kwargs)
        if isinstance(outs, (list, tuple)):
            m = outs[0]
        else:
            m = outs
        if isinstance(m, torch.Tensor):
            log_proba = F.log_softmax(m, dim=-1)
        elif isinstance(m, dict):
            log_proba = {
                k: F.log_softmax(v, dim=-1)
                for k, v in m.items()
            }
        else:
            raise ValueError("m must be a torch.Tensor or a dictionary")
        return log_proba

    def get_and_reset_regularization_loss(self) -> torch.Tensor:
        """
        Get the regularization loss as a sum of all the regularization losses of the layers. Then reset the
        regularization losses.

        :return: the regularization loss.
        :rtype: torch.Tensor
        """
        warnings.warn(
            "This method is deprecated and will be removed in the next version. Use get_regularization_loss instead.",
            DeprecationWarning
        )
        regularization_loss = torch.tensor(0.0, dtype=torch.float32, device=self.device)
        for layer in self.get_all_layers():
            if hasattr(layer, "get_and_reset_regularization_loss") and callable(layer.get_and_reset_regularization_loss):
                regularization_loss += layer.get_and_reset_regularization_loss()
        return regularization_loss





from copy import deepcopy
from typing import Dict, List, NamedTuple, Optional

# from mlagents_envs.side_channel.environment_parameters_channel import EnvironmentParametersChannel

from .buffers import ReplayBuffer


class CompletionCriteria(NamedTuple):
    """
    Completion criteria for a lesson.
    """
    measure: str
    min_lesson_length: int
    threshold: float

    @staticmethod
    def default_criteria() -> 'CompletionCriteria':
        return CompletionCriteria(measure='Rewards', min_lesson_length=1, threshold=0.9)


class Lesson:
    UNPICKLABLE_ATTRIBUTES = ['_teacher', '_channel']

    def __init__(
            self,
            name,
            channel,  #: EnvironmentParametersChannel,
            params: Dict[str, float],
            completion_criteria: CompletionCriteria = CompletionCriteria.default_criteria(),
            teacher=None,
            teacher_strength : Optional[float] = None
    ):
        # assert teacher is None or hasattr(teacher, 'buffer'), 'Teacher must have a replay buffer.'
        self.name = name
        self.completion_criteria = completion_criteria
        self._channel = channel
        self.params = params
        self._num_iterations = 0
        self._teacher = teacher
        self.teacher_strength = teacher_strength
        self._result = None

    @property
    def teacher(self):
        return self._teacher

    @teacher.setter
    def teacher(self, teacher):
        self._teacher = teacher

    @property
    def teacher_buffer(self) -> Optional[ReplayBuffer]:
        """
        Returns the replay buffer for the lesson.
        """
        if self._teacher is not None:
            buffer = self._teacher.buffer
            assert isinstance(buffer, ReplayBuffer), 'Teacher must have a replay buffer.'
            return self._teacher.buffer
        return None

    @property
    def is_completed(self):
        """
        Returns True if the lesson is completed, False otherwise.
        """
        return self._result is not None

    @property
    def channel(self):
        return self._channel

    @channel.setter
    def channel(self, channel):
        self._channel = channel

    def __getstate__(self):
        state = {
            k: v
            for k, v in self.__dict__.items()
            if k not in self.UNPICKLABLE_ATTRIBUTES
        }
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    def set_result(self, result):
        self._result = result
        if self._teacher is not None:
            self._teacher.discard_buffer()

    def start(self):
        """
        Starts the lesson.
        """
        for key, value in self.params.items():
            self._channel.set_float_parameter(key, value)

    def on_iteration_end(self, metrics: Dict[str, float]) -> bool:
        """
        Called when an iteration ends.
        """
        self._num_iterations += 1
        return self.check_completion_criteria(metrics)

    def check_completion_criteria(self, metrics: Dict[str, float]) -> bool:
        """
        Checks if the lesson is completed.
        """
        if self.completion_criteria.min_lesson_length > self._num_iterations:
            return False
        result = metrics[self.completion_criteria.measure]
        if result >= self.completion_criteria.threshold:
            self.set_result(result)
        return self.is_completed


class CurriculumEndIterationOutput(NamedTuple):
    """
    Output of the curriculum when the end of an iteration is reached.
    """
    messages: Dict[str, str]
    lesson_completed: bool


class Curriculum:
    def __init__(self, name: str = "Curriculum", description: str = "", lessons: List[Lesson] = None):
        self.name = name
        self.description = description
        self._lessons = [] if lessons is None else lessons
        self._current_lesson_idx = 0

    @property
    def is_completed(self) -> bool:
        """
        Returns True if the curriculum is completed, False otherwise.
        """
        return self._current_lesson_idx >= len(self._lessons)

    @property
    def current_lesson(self) -> Optional[Lesson]:
        """
        Returns the current lesson.
        """
        if self.is_completed:
            return None
        return self._lessons[self._current_lesson_idx]

    @property
    def map_repr(self) -> Dict[str, str]:
        progress = f"{100 * self._current_lesson_idx / (len(self._lessons)-1):.1f} %"
        if self.is_completed:
            return {self.name: f'(lesson: completed) 100%'}
        return {self.name: f'(lesson: {self.current_lesson.name}) {progress}'}

    @property
    def teachers(self):
        return [lesson.teacher for lesson in self._lessons]

    @property
    def channels(self):
        return [lesson.channel for lesson in self._lessons]

    @property
    def lessons(self):
        return self._lessons

    def __getitem__(self, item) -> Lesson:
        if isinstance(item, int):
            return self._lessons[item]
        elif isinstance(item, str):
            return next(filter(lambda l: l.name == item, self._lessons))
        else:
            raise TypeError(f"Invalid type for curriculum indexing: {type(item)}")

    def add_lesson(self, lesson: Lesson):
        self._lessons.append(lesson)

    def __str__(self):
        map_repr = self.map_repr
        return f'{self.name} {map_repr[self.name]}'

    def __repr__(self):
        return str(self)

    @property
    def teacher_buffer(self) -> Optional[ReplayBuffer]:
        """
        Returns the current teacher buffer.
        """
        if self.is_completed:
            return None
        return self._lessons[self._current_lesson_idx].teacher_buffer

    def on_iteration_start(self):
        """
        Called when an iteration starts.
        """
        if not self.is_completed:
            self.current_lesson.start()

    def on_iteration_end(self, metrics: Dict[str, float]) -> CurriculumEndIterationOutput:
        """
        Called when an iteration ends.
        """
        lesson_is_completed = False
        if not self.is_completed:
            self.current_lesson.on_iteration_end(metrics)
            if self.current_lesson.is_completed:
                self._current_lesson_idx += 1
                lesson_is_completed = True
        return CurriculumEndIterationOutput(messages=self.map_repr, lesson_completed=lesson_is_completed)

    def update_teachers(self, teachers: List):
        assert len(teachers) == len(self._lessons), 'Number of teachers must match number of lessons.'
        for lesson, teacher in zip(self._lessons, teachers):
            lesson.teacher = teacher

    def update_channels(self, channels: List):
        assert len(channels) == len(self._lessons), 'Number of channels must match number of lessons.'
        for lesson, channel in zip(self._lessons, channels):
            lesson._channel = channel

    def update_teachers_and_channels(self, other: 'Curriculum'):
        self.update_teachers(other.teachers)
        self.update_channels(other.channels)





import collections.abc
import hashlib
import pickle
from typing import Callable, Dict, List, Any, Union, Sequence, Type, Tuple

import torch


def get_meta_name(params: Dict[str, Any]):
    meta_name = f""
    keys = sorted(list(params.keys()))
    for k in keys:
        meta_name += f"{k}-{params[k]}_"
    return meta_name[:-1]


def hash_params(params: Dict[str, Any]):
    """
    Hash the parameters to get a unique and persistent id.

    Note: This is the old version of hash_params. It is kept for compatibility with old code.
        Please use hash_dict instead which is more general and offers more options.

    :param params: The parameters to hash.

    :return: The hash of the parameters.
    """
    return int(hashlib.md5(get_meta_name(params).encode('utf-8')).hexdigest(), 16)


def get_meta_str(__obj: Any) -> str:
    """
    Get the meta string of an object. The meta string is a string representation of the object
    that can be used as a file name. All mappings are sorted by keys before being converted to
    strings.

    :param __obj: The object to get the meta string of.

    :return: The meta string of the object.

    Examples:
        >>> get_meta_str(1)
        '1'
        >>> get_meta_str([1, 2, 3])
        '1_2_3'
        >>> get_meta_str({"a": 1, "b": 2})
        'a-1_b-2'
        >>> get_meta_str([{"b": 2, "a": 1}, {1: 2, 3: 4}])
        'a-1_b-2_1-2_3-4'
        >>> class CustomObject:
        ...     def __repr__(self):
        ...         return "my_repr"
        ... get_meta_str([{"b": 2, "a": 1}, {1: 2, 3: 4}, 5, 6, 7, CustomObject()])
        'a-1_b-2_1-2_3-4_5_6_7_my_repr'
    """
    meta_name = f""
    if isinstance(__obj, collections.abc.Mapping):
        keys = sorted(list(__obj.keys()))
        meta_name = '_'.join([f"{k}-{get_meta_str(__obj[k])}" for k in keys])
    elif isinstance(__obj, collections.abc.Iterable):
        meta_name = '_'.join([get_meta_str(k) for k in __obj])
    else:
        meta_name = str(repr(__obj))
    return meta_name


def hash_meta_str(
        __obj: Any,
        hash_mth: str = "md5",
        out_type: str = "hex"
) -> Union[str, int]:
    """
    Hash an object to get a unique and persistent id. The hash is computed by hashing the
    string representation of the entry. The string representation is obtained using the function
    `get_meta_str`.

    :param __obj: The object to hash.
    :param hash_mth: The hash method to use. Must be in hashlib.algorithms_available. Default is "md5".
    :param out_type: The type of the output. Must be in ["hex", "int"]. Default is "hex".

    :return: The hash of the object.
    """
    available_out_type = ["hex", "int"]
    import hashlib

    if hash_mth not in hashlib.algorithms_available:
        raise ValueError(f"hash_mth must be in {hashlib.algorithms_available}")
    hash_func = getattr(hashlib, hash_mth, hashlib.md5)
    if out_type == "hex":
        return hash_func(get_meta_str(__obj).encode('utf-8')).hexdigest()
    elif out_type == "int":
        return int(hash_func(get_meta_str(__obj).encode('utf-8')).hexdigest(), 16)
    raise ValueError(f"out_type must be in {available_out_type}")


def save_params(params: Dict[str, Any], save_path: str):
    """
    Save the parameters in a file.

    :param save_path: The path to save the parameters.
    :param params: The parameters to save.

    :return: The path to the saved parameters.
    """
    pickle.dump(params, open(save_path, "wb"))
    return save_path


def get_all_params_combinations(params_space: Dict[str, Any]) -> List[Dict[str, Any]]:
    """
    Get all possible combinations of parameters.

    :param params_space: Dictionary of parameters.

    :return: List of dictionaries of parameters.
    """
    import itertools
    # get all the combinaison of the parameters
    all_params = list(params_space.keys())
    all_params_values = list(params_space.values())
    all_params_combinaison = list(map(lambda x: list(x), list(itertools.product(*all_params_values))))

    # create a list of dict of all the combinaison
    all_params_combinaison_dict = list(map(lambda x: dict(zip(all_params, x)), all_params_combinaison))
    return all_params_combinaison_dict


def list_of_callable_to_sequential(callable_list: List[Callable]) -> torch.nn.Sequential:
    """
    Convert a list of callable to a list of modules.

    :param callable_list: List of callable.

    :return: List of modules.
    """
    from neurotorch.transforms.wrappers import CallableToModuleWrapper
    return torch.nn.Sequential(
        *[
            c if isinstance(c, torch.nn.Module) else CallableToModuleWrapper(c)
            for c in callable_list
        ]
    )


def sequence_get(__sequence: Sequence, idx: int, default: Any = None) -> Any:
    try:
        return __sequence[idx]
    except IndexError:
        return default


def list_insert_replace_at(__list: List, idx: int, value: Any):
    """
    Insert a value at a specific index. If there is already a value at this index, replace it.

    :param __list: The list to modify.
    :param idx: The index to insert the value.
    :param value: The value to insert.
    """
    if idx < len(__list):
        __list[idx] = value
    else:
        __list.extend([None] * (idx - len(__list)))
        __list.append(value)


def unpack_out_hh(out):
    """
    Unpack the output of a recurrent network.

    :param out: The output of a recurrent network.

    :return: The output of the recurrent network with the hidden state.
                If there is no hidden state, consider it as None.
    """
    out_tensor, hh = None, None
    if isinstance(out, (tuple, list)):
        if len(out) == 2:
            out_tensor, hh = out
        elif len(out) == 1:
            out_tensor = out[0]
        elif len(out) > 2:
            out_tensor, *hh = out
    else:
        out_tensor = out

    return out_tensor, hh


def unpack_x_hh_y(__input) -> Tuple[Any, Any, Any]:
    """
    Unpack the input, hidden state and target of a recurrent network.

    :param __input: The input to unpack.

    :return: The input of the recurrent network with the hidden state and the target.
                If there is no hidden state, consider it as None.
    """
    x, hh, y = None, None, None
    if isinstance(__input, (tuple, list)):
        if len(__input) == 3:
            x, hh, y = __input
        elif len(__input) == 2:
            x, y = __input
        elif len(__input) > 3:
            x, *hh, y = __input
        elif len(__input) == 1:
            x = __input[0]
    else:
        x = __input

    return x, hh, y


def unpack_tuple(
        x: Any,
        expected_length: int,
        fill_value: Any = None,
        fill_method: str = "middle",
        aggregate_method: str = "middle",
        aggregate_type: Type = tuple,
) -> tuple:
    """
    Unpack a tuple of a specific length. If the tuple has a different length, one of the following
    output is returned:
        - If the input 'x' has the expected length, the input 'x' unpacked is returned.
        - If the input 'x' has a length lower than the expected length, the input 'x' unpacked is
            returned with 'fill_value' appended to it using the 'fill_method'.
        - If the input 'x' has a length greater than the expected length, the input 'x' unpacked is
            returned with the elements packed together using the 'aggregate_method'.

    :param x: The tuple to unpack.
    :type x: Any
    :param expected_length: The expected length of the tuple.
    :type expected_length: int
    :param fill_value: The value to fill the tuple with if the length of the tuple is lower than the
        expected length. Default is None.
    :type fill_value: Any
    :param fill_method: The method to use to fill the tuple with 'fill_value's if the length of the tuple is lower than
        the expected length. Must be in ["left", "right", "middle"]. Default is "middle".
    :type fill_method: str
    :param aggregate_method: The method to use to aggregate the tuple if the length of the tuple is
        greater than the expected length. Must be in ["left", "right", "middle"]. Default is "middle".
    :type aggregate_method: str
    :param aggregate_type: The type of the aggregated items if the length of the tuple is greater than the
        expected length. Generally one of [tuple, list]. Default is tuple.

    :return: The unpacked tuple.
    :rtype: tuple

    TODO: Optimize this function.
    """
    output = [fill_value for _ in range(expected_length)]
    x_length = len(x)
    if x_length == expected_length:
        output = tuple(x)
    elif x_length < expected_length:
        if fill_method == "left":
            output[-x_length:] = list(x)
        elif fill_method == "right":
            output[:x_length] = list(x)
        elif fill_method == "middle":
            indexes = list(range(expected_length))
            n_fill = expected_length - x_length
            idx_start_fill = int(x_length / 2)
            idx_end_fill = idx_start_fill + n_fill - 1
            x_indexes = list(set(indexes) - set(range(idx_start_fill, idx_end_fill + 1)))
            for i, idx in enumerate(x_indexes):
                output[idx] = x[i]
        else:
            raise ValueError(f"fill_method must be in ['left', 'right', 'middle']. Got {fill_method}.")
    elif x_length > expected_length:
        n_aggregate = x_length - expected_length + 1
        if aggregate_method == "left":
            output = [aggregate_type(x[:n_aggregate])] + list(x[n_aggregate:])
        elif aggregate_method == "right":
            output = list(x[:-n_aggregate]) + [aggregate_type(x[-n_aggregate:])]
        elif aggregate_method == "middle":
            idx_start_aggregate = int(n_aggregate / 2)
            idx_end_aggregate = idx_start_aggregate + n_aggregate
            output = (
                    list(x[:idx_start_aggregate])
                    + [aggregate_type(x[idx_start_aggregate:idx_end_aggregate])]
                    + list(x[idx_end_aggregate:])
            )
        else:
            raise ValueError(f"aggregate_method must be in ['left', 'right', 'middle']. Got {aggregate_method}.")
    return tuple(output)


def unpack_singleton_dict(x: dict) -> Any:
    """
    Unpack a dictionary with a single key and value. If the dict has more than one key, a ValueError is raised.
    :param x:
    :return:
    """
    if len(x) > 1:
        raise ValueError("x must have a length of zero or one.")
    elif len(x) == 0:
        return None
    return x[list(x.keys())[0]]


def maybe_unpack_singleton_dict(x: Union[dict, Any]) -> Any:
    """
    Accept a dict or any other type. If x is a dict with one key and value, the singleton is unpacked. Otherwise, x is
    returned without being changed.
    :param x:
    :return:
    """
    if isinstance(x, dict) and len(x) <= 1:
        return unpack_singleton_dict(x)
    return x


def mapping_update_recursively(d, u):
    """
    from https://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth

    :param d: mapping item that wil be updated
    :param u: mapping item updater

    :return: updated mapping recursively
    """
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = mapping_update_recursively(d.get(k, {}), v)
        else:
            d[k] = v
    return d



from typing import Union, Dict, Any, Tuple, Type, Optional

import torch

from .spikes_decoders import MeanConv
from .spikes_encoders import SpikesEncoder
from ..modules.layers import SpyLIFLayer, LIFLayer, ALIFLayer
from ..modules.base import BaseModel


class SpikesAutoEncoder(BaseModel):

    @staticmethod
    def _default_encoder_type() -> Type[Union[LIFLayer, SpyLIFLayer, ALIFLayer]]:
        return SpyLIFLayer

    def __init__(
            self,
            n_units: int,
            n_encoder_steps: int,
            encoder_type: Optional[Type[Union[LIFLayer, SpyLIFLayer, ALIFLayer]]] = None,
            spikes_encoder: Optional[torch.nn.Module] = None,
            spikes_decoder: Optional[torch.nn.Module] = None,
            **kwargs
    ):
        if encoder_type is not None and spikes_encoder is not None:
            raise ValueError(
                "If encoder_type is provided, spikes_encoder must be None and vice versa."
            )
        kwargs.setdefault("name", "SpikesAutoEncoder")
        super().__init__(n_units, n_units, **kwargs)
        if encoder_type is None:
            encoder_type = self._default_encoder_type()
        self.encoder_type = encoder_type
        self.n_units = n_units
        self.n_encoder_steps = n_encoder_steps
        if spikes_encoder is None:
            spikes_encoder = self._create_default_encoder()
        self.spikes_encoder = spikes_encoder
        if spikes_decoder is None:
            spikes_decoder = self._create_default_decoder()
        self.spikes_decoder = spikes_decoder

    @BaseModel.device.setter
    def device(self, device: torch.device):
        self.spikes_encoder.to(device)
        self.spikes_decoder.to(device)
        BaseModel.device.fset(self, device)

    def _create_default_encoder(self):
        return SpikesEncoder(
            n_steps=self.n_encoder_steps,
            n_units=self.n_units,
            spikes_layer_type=self.encoder_type,
            device=self.device,
            dt=self.kwargs.pop("dt", 1e-3),
            **self.kwargs,
        )

    def _create_default_decoder(self):
        return MeanConv(
            self.n_encoder_steps,
            alpha=2.0,
            learn_alpha=True,
            learn_kernel=True,
            activation=torch.nn.Hardtanh(
                min_val=self.kwargs.get("hard_tanh.min_val", 0.0),
                max_val=self.kwargs.get("hard_tanh.max_val", 1.0)
            ),
        )

    def encode(self, x: torch.Tensor) -> torch.Tensor:
        return self.spikes_encoder(x)

    def decode(self, x: torch.Tensor):
        return self.spikes_decoder(x)

    def forward(self, x: torch.Tensor, **kwargs):
        x = self._to_device_transform(x)
        return self.decode(self.encode(x))

    def get_prediction_trace(
            self, inputs: Union[Dict[str, Any], torch.Tensor],
            **kwargs
    ) -> Union[Dict[str, torch.Tensor], torch.Tensor]:
        raise NotImplementedError()

    def get_raw_prediction(
            self,
            inputs: torch.Tensor,
            *args,
            **kwargs
    ) -> Union[Tuple[Any, Any, Any], Tuple[Any, Any], Any]:
        return self.forward(inputs)

    def get_and_reset_regularization_loss(self) -> torch.Tensor:
        """
        Get the regularization loss as a sum of all the regularization losses of the layers. Then reset the
        regularization losses.

        :return: the regularization loss.
        """
        regularization_loss = torch.tensor(0.0, dtype=torch.float32, device=self.device)
        for layer in [self.spikes_encoder, self.spikes_decoder]:
            if hasattr(layer, "get_and_reset_regularization_loss") and callable(
                    layer.get_and_reset_regularization_loss
            ):
                regularization_loss += layer.get_and_reset_regularization_loss()
        return regularization_loss








